# Copyright (c) 2015-2017 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Basic test Class and elements for testing Cisco Nexus platforms.

Most classes in this file do not contain test cases but instead
provide common methods for other classes to utilize. This class
provides the basic methods needed to drive a create or delete
port request thru to the ssh or restapi driver. It verifies the
final content of the data base and verifies what data the
Drivers sent out.  There also exists another 'base' class
specifically for Replay testing.
"""

import collections
import mock
from operator import attrgetter
from oslo_config import cfg
import re
import six
import testtools

from networking_cisco import backwards_compatibility as bc
from networking_cisco.backwards_compatibility import constants as p_const
from networking_cisco.backwards_compatibility import ml2_api as api

from networking_cisco.ml2_drivers.nexus import (
    constants as const)
from networking_cisco.ml2_drivers.nexus import (
    nexus_helpers as nexus_help)
from networking_cisco.ml2_drivers.nexus import (
    nexus_network_driver)
from networking_cisco.ml2_drivers.nexus import (
    nexus_restapi_network_driver)
from networking_cisco.ml2_drivers.nexus import (
    nexus_restapi_snippets as rest_snipp)
from networking_cisco.ml2_drivers.nexus import exceptions
from networking_cisco.ml2_drivers.nexus import mech_cisco_nexus
from networking_cisco.ml2_drivers.nexus import nexus_db_v2
from networking_cisco.ml2_drivers.nexus import trunk

from neutron.tests.unit import testlib_api

from networking_cisco.tests import base as nc_base


CONF = cfg.CONF


# Static variables used in testing
NEXUS_IP_ADDRESS_1 = '1.1.1.1'
NEXUS_IP_ADDRESS_2 = '2.2.2.2'
NEXUS_IP_ADDRESS_3 = '3.3.3.3'
NEXUS_IP_ADDRESS_DUAL = '4.4.4.4'
NEXUS_IP_ADDRESS_DUAL2 = '5.5.5.5'
NEXUS_IP_ADDRESS_6 = '6.6.6.6'
NEXUS_IP_ADDRESS_7 = '7.7.7.7'
NEXUS_IP_ADDRESS_8 = '8.8.8.8'

HOST_NAME_1 = 'testhost1'
HOST_NAME_2 = 'testhost2'
HOST_NAME_PC = 'testpchost'
HOST_NAME_DUAL = 'testdualhost'
HOST_NAME_3 = 'testhost3'
HOST_NAME_4 = 'testhost4'
HOST_NAME_5 = 'testhost5'
HOST_NAME_6 = 'testhost6'
HOST_NAME_UNUSED = 'unused'
HOST_NAME_Baremetal = 'baremetal'

INSTANCE_1 = 'testvm1'
INSTANCE_2 = 'testvm2'
INSTANCE_PC = 'testpcvm'
INSTANCE_DUAL = 'testdualvm'

NEXUS_BAREMETAL_PORT_1 = 'Ethernet 1/10'
NEXUS_BAREMETAL_PORT_2 = 'Ethernet 1/20'
NEXUS_PORT_1 = 'ethernet:1/10'
NEXUS_PORT_2 = 'ethernet:1/20'
NEXUS_PORT_3 = 'ethernet:1/30'
NEXUS_DUAL1 = 'ethernet:1/3'
NEXUS_DUAL2 = 'ethernet:1/2'
NEXUS_PORTCHANNELS = 'portchannel:2'
NEXUS_DUAL = 'ethernet:1/3,portchannel:2'
NEXUS_DUAL_2 = '1/2,1/3'

VLAN_ID_1 = 267
VLAN_ID_2 = 265
VLAN_ID_PC = 268
VLAN_ID_DUAL = 269

VXLAN_ID = 70000
NO_VXLAN_ID = 0

MCAST_GROUP = '255.1.1.1'

PHYSNET = 'physnet1'
NETID = 999
PORT_ID = 'fakePortID'

DEVICE_OWNER_BAREMETAL = 'baremetal:none'
DEVICE_OWNER_COMPUTE = 'compute:test'
DEVICE_OWNER_DHCP = bc.constants.DEVICE_OWNER_DHCP
DEVICE_OWNER_ROUTER_HA_INTF = bc.constants.DEVICE_OWNER_ROUTER_HA_INTF
DEVICE_OWNER_ROUTER_INTF = bc.constants.DEVICE_OWNER_ROUTER_INTF
DEVICE_OWNER_ROUTER_GW = bc.constants.DEVICE_OWNER_ROUTER_GW

NEXUS_SSH_PORT = '22'
PORT_STATE = bc.constants.PORT_STATUS_ACTIVE
NETWORK_TYPE = 'vlan'
VLAN_TYPE_TRUNK = 'trunk'
VLAN_TYPE_NATIVE = 'native'

NORMAL_VNIC = u'normal'
BAREMETAL_VNIC = u'baremetal'

CONNECT_ERROR = 'Unable to connect to Nexus'

GET_NEXUS_TYPE_RESPONSE = {
    "totalCount": "1",
    "imdata": [
        {
            "eqptCh": {
                "attributes": {
                    "descr": "Nexus9000 C9396PX Chassis",
                }
            }
        }
    ]
}

GET_INTERFACE_RESPONSE = {
    "totalCount": "1",
    "imdata": [
        {
            "l1PhysIf": {
                "attributes": {
                    "mode": "trunk",
                    "trunkVlans": ""
                }
            }
        }
    ]
}

GET_INTERFACE_PCHAN_RESPONSE = {
    "totalCount": "1",
    "imdata": [
        {
            "pcAggrIf": {
                "attributes": {
                    "mode": "trunk",
                    "trunkVlans": ""
                }
            }
        }
    ]
}


GET_NO_PORT_CH_RESPONSE = {
    "totalCount": "3",
    "imdata": [
    ]
}

POST = 0
DELETE = 1

## Test snippets used to verify nexus command output
RESULT_ADD_VLAN = """configure\>\s+\<vlan\>\s+\
<vlan-id-create-delete\>\s+\<__XML__PARAM_value\>{0}"""

RESULT_ADD_VLAN_VNI = """configure\>\s+\<vlan\>\s+\
<vlan-id-create-delete\>\s+\<__XML__PARAM_value\>{0}[\x20-\x7e]+\
\s+[\x20-\x7e]+\s+\<vn-segment\>\s+\<vlan-vnsegment\>{1}"""

RESULT_ADD_INTERFACE = """\<{0}\>\s+\<interface\>\
{1}\<\/interface\>\s+[\x20-\x7e]+\s+\<switchport\>\s+\<trunk\>\s+\
\<allowed\>\s+\<vlan\>\s+\<add\>\s+\<vlan_id\>{2}"""

RESULT_ADD_NATIVE_INTERFACE = """\<{0}\>\s+\<interface\>\
{1}\<\/interface\>\s+[\x20-\x7e]+\s+\<switchport\>\s+\<trunk\>\s+\
\<native\>\s+\<vlan\>\s+\<vlan_id\>{2}"""

RESULT_ADD_NVE_INTERFACE = """\<interface\>\s+\
\<nve\>nve{0}\<\/nve\>\s+[\x20-\x7e]+\s+\<member\>member vni {1} \
mcast-group {2}"""

RESULT_INTERFACE = """\<{0}\>\s+\<interface\>\
{1}\<\/interface\>\s+[\x20-\x7e]+\s+\<switchport\>\s+\<trunk\>\s+\
\<allowed\>\s+\<vlan\>\s+\<vlan_id\>{2}"""

RESULT_DEL_VLAN = """\
\<no\>\s+\<vlan\>\s+<vlan-id-create-delete\>\
\s+\<__XML__PARAM_value\>{0}"""

RESULT_DEL_INTERFACE = """\
\<{0}\>\s+\<interface\>{1}\<\/interface\>\s+\
[\x20-\x7e\s]+\<switchport\>\s+\<trunk\>\s+\
\<allowed\>\s+\<vlan\>\s+\<remove\>\s+\<vlan\>{2}"""

RESULT_DEL_NATIVE_INTERFACE = """\
\<{0}\>\s+\<interface\>{1}\<\/interface\>\s+\
[\x20-\x7e\s]+\<no\>\s+\<switchport\>\s+\<trunk\>\s+\
\<native\>"""

RESULT_DEL_NVE_INTERFACE = """\<interface\>\s+\
\<nve\>nve{0}\<\/nve\>\s+[\x20-\x7e]+\s+\<member\>no member vni {1}"""


class FakeNetworkContext(object):
    """Network context for testing purposes only."""

    def __init__(self, segment_id, nw_type, mcast_group='physnet1',
                 netid=None):

        if not netid:
            netid = NETID
        self._network_segments = {api.SEGMENTATION_ID: segment_id,
                                  api.ID: netid,
                                  api.NETWORK_TYPE: nw_type,
                                  api.PHYSICAL_NETWORK: mcast_group}

    @property
    def network_segments(self):
        return self._network_segments

    @property
    def current(self):
        return self._network_segments


class FakePortContext(object):

    """Port context for testing purposes only."""

    def __init__(self, device_id, host_name, device_owner,
                 network_context, bottom_segment=None,
                 profile=None, vnic_type=u'normal',
                 dns_name=None, netid=None):
        if profile is None:
            profile = []
        if not netid:
            netid = NETID
        self._set_port(device_id, host_name, device_owner,
                       profile, vnic_type, dns_name, netid)
        self._port_orig = None
        self._network = network_context
        if network_context:
            self._segment = network_context.network_segments
            self.segments_to_bind = [network_context.network_segments]
        if bottom_segment is None:
            self._bottom_segment = None
        else:
            self._bottom_segment = bottom_segment.network_segments
        self._segments_to_bind = None

    def _set_port(self, device_id, host_name, device_owner,
                  profile=None, vnic_type=u'normal',
                  dns_name=None, netid=None):

        self._port = {
            'status': PORT_STATE,
            'device_id': device_id,
            'device_owner': device_owner,
            api.ID: PORT_ID,
            'dns_name': dns_name,
            'network_id': netid,
            bc.portbindings.HOST_ID: host_name,
            bc.portbindings.VNIC_TYPE: vnic_type,
            bc.portbindings.PROFILE: profile,
            bc.portbindings.VIF_TYPE: bc.portbindings.VIF_TYPE_OVS
        }

    def set_orig_port(self, device_id, host_name, device_owner,
                      profile=None, vnic_type=u'normal',
                      dns_name=None, netid=None):

        self._port_orig = {
            'status': PORT_STATE,
            'device_id': device_id,
            'device_owner': device_owner,
            api.ID: PORT_ID,
            'dns_name': dns_name,
            'network_id': netid,
            bc.portbindings.HOST_ID: host_name,
            bc.portbindings.VNIC_TYPE: vnic_type,
            bc.portbindings.PROFILE: profile,
            bc.portbindings.VIF_TYPE: bc.portbindings.VIF_TYPE_OVS
        }

    @property
    def current(self):
        return self._port

    @property
    def original(self):
        return self._port_orig

    @property
    def network(self):
        return self._network

    @property
    def top_bound_segment(self):
        return self._segment

    @property
    def bottom_bound_segment(self):
        return self._bottom_segment

    @property
    def original_top_bound_segment(self):
        return None

    @property
    def original_bottom_bound_segment(self):
        return None

    def continue_binding(self, segment_id, next_segments_to_bind):
        pass

    def set_binding(self, segment_id, vif_type, vif_details,
                    status=None):
        pass

    def allocate_dynamic_segment(self, segment):
        pass

    def _prepare_to_bind(self, segments_to_bind):
        self._segments_to_bind = segments_to_bind
        self._new_bound_segment = None
        self._next_segments_to_bind = None

NEXUS_CONF_TEMPLATE = """
[ml2_mech_cisco_nexus:%(ip_addr)s]
ssh_port=%(ssh_port)s
username=admin
password=password
physnet=%(physnet)s
"""


class FakeUnbindPortContext(FakePortContext):
    """Port context used during migration to unbind port."""
    @property
    def top_bound_segment(self):
        return None

    @property
    def bottom_bound_segment(self):
        return None

    @property
    def original_top_bound_segment(self):
        return self._segment

    @property
    def original_bottom_bound_segment(self):
        return self._bottom_segment


class TestCiscoNexusBaseResults(object):

    """Unit tests driver results for Cisco ML2 Nexus."""

    test_results = {}

    def get_test_results(self, name):
        if name in self.test_results:
            return self.test_results[name]
        else:
            return None


HOST_MAPPING_CONFIG_FILE = """
[ml2_mech_cisco_nexus:1.1.1.1]
username=admin
password=mySecretPassword
compute1=1/1
compute2=1/2
compute5=1/3,1/4

[ml2_mech_cisco_nexus:2.2.2.2]
username=admin
password=mySecretPassword
compute3=1/1
compute4=1/2
compute5=port-channel:20,port-channel:30
"""

DICT_MAPPING_CONFIG_FILE = """
[ml2_mech_cisco_nexus:1.1.1.1]
username=admin
password=mySecretPassword
nve_src_intf=2
physnet=physnet1
host_ports_mapping=compute1:[1/1, 1/2],
                   compute2:[1/3],
                   compute3:[port-channel30]
"""


class TestCiscoNexusPluginHostMapping(testlib_api.SqlTestCase):

    def setUp(self):
        super(TestCiscoNexusPluginHostMapping, self).setUp()
        cfg.CONF.clear()
        self.cisco_mech_driver = mech_cisco_nexus.CiscoNexusMechanismDriver()

    def _assert_expected_mappings_get_added(self, expected_host_map_data):
        """Assert all expected mappings now exist and there aren't any
        unexpected mappings
        """
        # Assert there are currently no mappings
        self.assertRaises(exceptions.NexusHostMappingNotFound,
                          nexus_db_v2.get_all_host_mappings)

        # Call initialize host mapping function
        self.cisco_mech_driver._initialize_host_port_mappings()

        mappings = nexus_db_v2.get_all_host_mappings()
        for hostmap in mappings:
            self.assertEqual(
                expected_host_map_data[hostmap.switch_ip][hostmap.if_id],
                hostmap.host_id)
            self.assertEqual(0, hostmap.ch_grp)
            self.assertTrue(hostmap.is_static)
            # Remove this mapping from the expected_host_map_data
            del expected_host_map_data[hostmap.switch_ip][hostmap.if_id]
        # Assert we've seen and removed all the expected host mappings
        for ip, mappings in expected_host_map_data.items():
            self.assertEqual({}, mappings)

    def test__initialize_host_port_mappings(self):
        """Verify port-mapping table is configured correctly."""

        nc_base.load_config_file(HOST_MAPPING_CONFIG_FILE)
        expected_host_map_data = {
            '1.1.1.1': {
                'ethernet:1/1': 'compute1',
                'ethernet:1/2': 'compute2',
                'ethernet:1/3': 'compute5',
                'ethernet:1/4': 'compute5',
            },
            '2.2.2.2': {
                'ethernet:1/1': 'compute3',
                'ethernet:1/2': 'compute4',
                'port-channel:20': 'compute5',
                'port-channel:30': 'compute5',
            },
        }

        # Assert all expected mappings now exist and there aren't any
        # unexpected mappings
        self._assert_expected_mappings_get_added(expected_host_map_data)

    def test__initialize_host_port_mappings_with_dict(self):
        """Verify port-mapping table is configured correctly."""

        nc_base.load_config_file(DICT_MAPPING_CONFIG_FILE)
        expected_host_map_data = {
            '1.1.1.1': {
                'ethernet:1/1': 'compute1',
                'ethernet:1/2': 'compute1',
                'ethernet:1/3': 'compute2',
                'port-channel:30': 'compute3',
            },
        }

        # Assert all expected mappings now exist and there aren't any
        # unexpected mappings
        self._assert_expected_mappings_get_added(expected_host_map_data)

    @mock.patch.object(mech_cisco_nexus.CiscoNexusMechanismDriver,
                       '_initialize_host_port_mappings', autospec=True)
    def test_initialize_calls_init_host_mappings(self, mock_init_host):
        nc_base.load_config_file(HOST_MAPPING_CONFIG_FILE)
        self.cisco_mech_driver.initialize()
        mock_init_host.assert_called_once_with(self.cisco_mech_driver)


class TestCiscoNexusBase(testlib_api.SqlTestCase):
    """Feature Base Test Class for Cisco ML2 Nexus driver."""

    TestConfigObj = collections.namedtuple(
        'TestConfigObj',
        'nexus_ip_addr host_name nexus_port instance_id vlan_id vxlan_id '
        'mcast_group device_owner profile dns_name vnic_type')

    def mock_init(self):

        # This initializes interface responses to prevent
        # unnecessary noise to the results.

        data_xml = {'connect.return_value.get.return_value.data_xml':
                    'switchport trunk allowed vlan none'}
        self.mock_ncclient.configure_mock(**data_xml)

    def restapi_mock_init(self):

        # This initializes RESTAPI responses to prevent
        # unnecessary noise to the results.

        data_json = {'rest_get.side_effect':
                    self.get_side_effect}
        self.mock_ncclient.configure_mock(**data_json)

    def get_side_effect(self, action, ipaddr=None, body=None, headers=None):

        eth_path = 'api/mo/sys/intf/phys-'
        port_chan_path = 'api/mo/sys/intf/aggr-'

        if action == rest_snipp.PATH_GET_NEXUS_TYPE:
            return GET_NEXUS_TYPE_RESPONSE
        elif action in rest_snipp.PATH_GET_PC_MEMBERS:
            return GET_NO_PORT_CH_RESPONSE
        elif eth_path in action:
            return GET_INTERFACE_RESPONSE
        elif port_chan_path in action:
            return GET_INTERFACE_PCHAN_RESPONSE

        return {}

    def _make_vpc_list(self, from_in, to_in):

        new_list = []
        for x in range(from_in, (to_in + 1)):
            new_list.append(x)
        return new_list

    def _clear_port_dbs(self):
        nexus_db_v2.remove_all_nexusport_bindings()

    def _set_switch_state(self, port_cfg, state):

        # not baremetal transaction
        if port_cfg.nexus_ip_addr:
            self._cisco_mech_driver.set_switch_ip_and_active_state(
                port_cfg.nexus_ip_addr, state)
        else:    # baremetal transaction
            if port_cfg.profile:
                all_link_info = port_cfg.profile['local_link_information']
                for link_info in all_link_info:
                    ip_addr = link_info['switch_info']['switch_ip']
                    self._cisco_mech_driver.set_switch_ip_and_active_state(
                        ip_addr, state)

    def _get_ip_addrs(self, port_cfg):

        # not baremetal transaction
        if port_cfg.nexus_ip_addr:
            if port_cfg.host_name.startswith('baremetal'):
                maps = nexus_db_v2.get_host_mappings(port_cfg.host_name)
                ipaddrs = []
                for map in maps:
                    ipaddrs.append(map.switch_ip)
                return ipaddrs
            return [port_cfg.nexus_ip_addr]
        else:    # baremetal transaction
            ipaddrs = []
            all_link_info = port_cfg.profile['local_link_information']
            for link_info in all_link_info:
                ipaddrs.append(link_info['switch_info']['switch_ip'])
            return ipaddrs

    def _config_switch_cred(self, test_config_parts, ip_addr):
        if ip_addr not in test_config_parts:
            test_config_parts[ip_addr] = {}
            test_config_parts[ip_addr]['main'] = (
                NEXUS_CONF_TEMPLATE % {'ip_addr': ip_addr,
                                       'ssh_port': NEXUS_SSH_PORT,
                                       'physnet': PHYSNET})

    def setUp(self):
        """Sets up mock client, switch, and credentials dictionaries."""

        #Clear all configuration parsing
        CONF.clear()

        super(TestCiscoNexusBase, self).setUp()

        CONF.import_opt('api_workers', 'neutron.service')
        CONF.set_default('api_workers', 0)
        CONF.import_opt('rpc_workers', 'neutron.service')
        CONF.set_default('rpc_workers', 0)

        # Use a mock netconf or REST API client
        self.mock_ncclient = mock.Mock()
        if CONF.ml2_cisco.nexus_driver == 'restapi':
            mock.patch.object(
                nexus_restapi_network_driver.CiscoNexusRestapiDriver,
                '_import_client',
                return_value=self.mock_ncclient).start()
            self.mock_nxapi_client = mock.Mock()
            mock.patch.object(
                nexus_restapi_network_driver.CiscoNexusRestapiDriver,
                '_get_nxapi_client',
                return_value=self.mock_nxapi_client).start()
            self._verify_results = self._verify_restapi_results
        else:
            mock.patch.object(
                nexus_network_driver.CiscoNexusSshDriver,
                '_import_client',
                return_value=self.mock_ncclient).start()
            self._verify_results = self._verify_ssh_results

        original_get_switch_ips = (
            mech_cisco_nexus.CiscoNexusMechanismDriver.get_switch_ips)

        original_get_switch_host_mappings = (
            nexus_db_v2.get_switch_host_mappings)

        original_get_host_mappings = (
            nexus_db_v2.get_host_mappings)

        def new_get_switch_ips(self):
            switch_ips = original_get_switch_ips(self)
            switch_ips.sort()
            return switch_ips

        def new_get_switch_host_mappings(switch_ip):
            map = original_get_switch_host_mappings(switch_ip)
            map = sorted(map, key=attrgetter('if_id'))
            return map

        def new_get_host_mappings(host_id):
            map = original_get_host_mappings(host_id)
            map = sorted(map, key=attrgetter('switch_ip', 'if_id'))
            return map

        mock.patch.object(nexus_db_v2,
                         'get_switch_host_mappings',
                         new=new_get_switch_host_mappings).start()
        mock.patch.object(nexus_db_v2,
                         'get_host_mappings',
                         new=new_get_host_mappings).start()
        mock.patch.object(mech_cisco_nexus.CiscoNexusMechanismDriver,
                         'get_switch_ips',
                         new=new_get_switch_ips).start()
        test_config_parts = {}
        for name, config in self.test_configs.items():
            host_name = config.host_name
            nexus_port = config.nexus_port
            if not config.nexus_ip_addr:
                if not config.profile:
                    continue
                all_link_info = config.profile['local_link_information']
                for link_info in all_link_info:
                    ip_addr = link_info['switch_info']['switch_ip']
                    self._config_switch_cred(test_config_parts, ip_addr)
            else:
                ip_addr = config.nexus_ip_addr
                self._config_switch_cred(test_config_parts, ip_addr)

            if (host_name is not HOST_NAME_UNUSED and
               HOST_NAME_Baremetal not in host_name):
                if host_name in test_config_parts[ip_addr]:
                    test_config_parts[ip_addr][host_name].add(nexus_port)
                else:
                    test_config_parts[ip_addr][host_name] = set([nexus_port])
        test_config_file = ""
        for ip, subparts in test_config_parts.items():
            switch_config = subparts['main']
            for name, subpart in subparts.items():
                if name == "main":
                    continue
                switch_config += "%s=%s\n" % (name, ','.join(subpart))
            test_config_file += switch_config
        nc_base.load_config_file(test_config_file)

        self.mock_continue_binding = mock.patch.object(
            FakePortContext,
            'continue_binding').start()

        self.mock_get_dynamic_segment = mock.patch.object(
            bc.segments_db, 'get_dynamic_segment',
            return_value={'testkey': 'testvalue'}).start()

        mock.patch.object(
            trunk.NexusMDTrunkHandler, 'is_trunk_subport_baremetal',
            return_value=False).start()

        if CONF.ml2_cisco.nexus_driver == 'restapi':
            self.restapi_mock_init()
        else:
            self.mock_init()
        self._cisco_mech_driver = mech_cisco_nexus.CiscoNexusMechanismDriver()
        self._cisco_mech_driver.initialize()
        self._cfg_monitor = self._cisco_mech_driver.monitor
        self._cisco_mech_driver.driver.nexus_switches = (
            CONF.ml2_cisco.nexus_switches)
        self.addCleanup(self._clear_port_dbs)

    def _generate_port_context(self, port_config,
                               override_host_name=None,
                               override_netid=None,
                               unbind_port=False):
        """Returns port context from port_config."""

        if override_host_name:
            host_name = override_host_name
        else:
            host_name = port_config.host_name
        instance_id = port_config.instance_id
        vlan_id = port_config.vlan_id
        vxlan_id = port_config.vxlan_id
        mcast_group = port_config.mcast_group
        device_owner = port_config.device_owner
        profile = port_config.profile
        vnic_type = port_config.vnic_type
        dns_name = port_config.dns_name
        if override_netid:
            netid = override_netid
        else:
            netid = None

        network_context = FakeNetworkContext(vlan_id, NETWORK_TYPE, netid)
        if vxlan_id != NO_VXLAN_ID:
            vxlan_network_context = FakeNetworkContext(vxlan_id,
                const.TYPE_NEXUS_VXLAN, mcast_group, netid)
            if unbind_port:
                port_context = FakeUnbindPortContext(
                    instance_id, host_name, device_owner,
                    vxlan_network_context, network_context,
                    profile, vnic_type, dns_name, netid)
            else:
                port_context = FakePortContext(
                    instance_id, host_name, device_owner,
                    vxlan_network_context, network_context,
                    profile, vnic_type, dns_name, netid)
        else:
            if unbind_port:
                port_context = FakeUnbindPortContext(
                    instance_id, host_name, device_owner,
                    network_context, None,
                    profile, vnic_type, dns_name, netid)
            else:
                port_context = FakePortContext(
                    instance_id, host_name, device_owner,
                    network_context, None,
                    profile, vnic_type, dns_name, netid)

        return port_context

    def _bind_port(self, port_config, expect_success=True):
        """Tests creation of a virtual port."""

        port_context = self._generate_port_context(port_config)

        self.mock_set_binding = mock.patch.object(
            FakePortContext,
            'set_binding').start()

        port_context._prepare_to_bind(port_context)
        self._cisco_mech_driver.bind_port(port_context)

    def _create_port(self, port_config, override_netid=None):
        """Tests creation of a virtual port."""

        port_context = self._generate_port_context(
            port_config, override_netid=override_netid)

        self._cisco_mech_driver.create_port_postcommit(port_context)
        self._cisco_mech_driver.bind_port(port_context)
        self._cisco_mech_driver.update_port_precommit(port_context)
        self._cisco_mech_driver.update_port_postcommit(port_context)

        if nexus_help.is_baremetal(port_context.current):
            connections = self._cisco_mech_driver._get_port_connections(
                port_context.current, '')
        else:
            connections = self._cisco_mech_driver._get_port_connections(
                port_context.current, port_config.host_name)

        # for port_id in port_config.nexus_port.split(','):
        for switch_ip, intf_type, port, is_p_vlan, _ in connections:
            if switch_ip is not port_config.nexus_ip_addr:
                continue
            port_id = intf_type + ':' + port
            bindings = nexus_db_v2.get_nexusport_binding(
                           port_id,
                           port_config.vlan_id,
                           port_config.nexus_ip_addr,
                           port_config.instance_id)
            self.assertEqual(1, len(bindings))

    def _verify_restapi_results(self, driver_result):
        """Verifies correct entries sent to Nexus."""

        posts = 0
        deletes = 0
        for idx in range(0, len(driver_result)):
            if driver_result[idx][3] == POST:
                posts += 1
            else:
                deletes += 1
        self.assertEqual(
            posts,
            len(self.mock_ncclient.rest_post.mock_calls),
            "Unexpected driver post calls")
        self.assertEqual(
            deletes,
            len(self.mock_ncclient.rest_delete.mock_calls),
            "Unexpected driver delete calls")

        post_calls = self.mock_ncclient.rest_post.mock_calls
        del_calls = self.mock_ncclient.rest_delete.mock_calls
        posts = 0
        deletes = 0
        for idx in range(0, len(driver_result)):
            # assigned None to skip testing this one.
            if not driver_result[idx]:
                continue
            if driver_result[idx][3] == POST:
                test_it = post_calls[posts][1]
            else:
                test_it = del_calls[deletes][1]
            self.assertTrue(
                (driver_result[idx][0] ==
                    test_it[0]),
                "Expected Rest URI does not match")

            if driver_result[idx][1] is not None:
                self.assertTrue(
                    (driver_result[idx][1] ==
                        test_it[1]),
                    "Expected Nexus Switch ip does not match")

            if driver_result[idx][3] == POST:
                self.assertTrue(
                    (driver_result[idx][2] ==
                        test_it[2]),
                    "Expected Rest Body does not match")
                posts += 1
            else:
                deletes += 1

    def _delete_port(self, port_config):
        """Tests deletion of a virtual port."""
        port_context = self._generate_port_context(port_config)

        self._cisco_mech_driver.delete_port_precommit(port_context)
        self._cisco_mech_driver.delete_port_postcommit(port_context)

        if nexus_help.is_baremetal(port_context.current):
            connections = self._cisco_mech_driver._get_port_connections(
                port_context.current, '')
        else:
            connections = self._cisco_mech_driver._get_port_connections(
                port_context.current, port_config.host_name)

        # for port_id in port_config.nexus_port.split(','):
        for switch_ip, intf_type, port, is_p_vlan, _ in connections:
            if switch_ip is not port_config.nexus_ip_addr:
                continue
            port_id = intf_type + ':' + port
            with testtools.ExpectedException(
                    exceptions.NexusPortBindingNotFound):
                nexus_db_v2.get_nexusport_binding(
                    port_id,
                    port_config.vlan_id,
                    port_config.nexus_ip_addr,
                    port_config.instance_id)

    def _verify_ssh_results(self, driver_result):
        """Verifies correct entries sent to Nexus."""

        self.assertEqual(
            len(driver_result),
            self.mock_ncclient.connect.return_value.
            edit_config.call_count,
            "Unexpected driver count")

        for idx in range(0, len(driver_result)):
            self.assertIsNotNone(
                self.mock_ncclient.connect.
                return_value.edit_config.mock_calls[idx][2]['config'],
                "mock_data is None")
            # assign None to skip testing this one.
            if driver_result[idx]:
                self.assertIsNotNone(
                    re.search(driver_result[idx],
                        self.mock_ncclient.connect.return_value.
                        edit_config.mock_calls[idx][2]['config']),
                    "Expected result data not found")

    def _cfg_vPC_user_commands(self, nexus_ips, cmds):
        # Use commands provided by user instead of
        # sending BODY_ADD_PORT_CH_P2.  So
        # BODY_USER_CONF_CMDS will be sent instead.
        for sw_ip in nexus_ips:
            CONF.set_override(
                const.IF_PC, cmds,
                CONF.ml2_cisco.nexus_switches.get(sw_ip)._group)

    def _verify_nxapi_results(self, driver_result):
        """Verifies correct NXAPI entries sent to Nexus."""

        self.assertEqual(
            len(driver_result),
            self.mock_nxapi_client.rest_post.call_count,
            "Unexpected NXAPI driver count")

        for idx in range(0, len(driver_result)):
            if not driver_result[idx]:
                continue
            self.assertTrue(
                (driver_result[idx][0] ==
                    self.mock_nxapi_client.rest_post.mock_calls[idx][1][0]),
                "Expected Rest URI does not match")
            if driver_result[idx][1] is not None:
                self.assertTrue(
                    (driver_result[idx][1] ==
                        self.mock_nxapi_client.
                        rest_post.mock_calls[idx][1][1]),
                    "Expected Nexus Switch ip does not match")
            self.assertIsNotNone(
                self.mock_nxapi_client.rest_post.mock_calls[idx][1][2],
                "mock_data is None")
            self.assertIsNotNone(
                re.search(driver_result[idx][2],
                    self.mock_nxapi_client.rest_post.mock_calls[idx][1][2]),
                "Expected result data not found in NXAPI output")

    def _basic_create_verify_port_vlan(self, test_name, test_result,
                                       nbr_of_bindings=1,
                                       other_test=None):
        """Create port vlan and verify results."""

        if other_test is None:
            other_test = self.test_configs[test_name]

        # Configure port entry config which puts switch in inactive state
        self._create_port(other_test)

        port_cfg = other_test

        ipaddrs = self._get_ip_addrs(port_cfg)
        bindings_found = 0
        for ipaddr in ipaddrs:
            try:
                port_bindings = nexus_db_v2.get_nexusport_switch_bindings(
                    ipaddr)
                bindings_found += len(port_bindings)
                # Verify it's in the port binding data base
                # Add one to count for the reserved switch state entry
                # if replay is enabled
                if self._cisco_mech_driver.is_replay_enabled():
                    nbr_of_bindings += 1
            except exceptions.NexusPortBindingNotFound:
                pass
        self.assertEqual(nbr_of_bindings, bindings_found)

        # Make sure there is only a single attempt to configure.
        self._verify_results(test_result)

        # Clean all the ncclient mock_calls to clear exception
        # and other mock_call history.
        self.mock_ncclient.reset_mock()

    def _basic_delete_verify_port_vlan(self, test_name, test_result,
                                       nbr_of_bindings=0,
                                       nbr_of_mappings=0,
                                       other_test=None):
        """Create port vlan and verify results."""

        if other_test is None:
            other_test = self.test_configs[test_name]

        self._delete_port(other_test)

        # Verify port binding has been removed
        # Verify failure stats is not reset and
        # verify no driver transactions have been sent
        port_cfg = other_test
        ipaddrs = self._get_ip_addrs(port_cfg)
        bindings_found = 0
        for ipaddr in ipaddrs:
            try:
                port_bindings = nexus_db_v2.get_nexusport_switch_bindings(
                    ipaddr)
                bindings_found += len(port_bindings)
                if self._cisco_mech_driver.is_replay_enabled():
                    # Add one for the reserved switch state entry
                    nbr_of_bindings += 1
            except exceptions.NexusPortBindingNotFound:
                pass
        self.assertEqual(nbr_of_bindings, bindings_found)

        port_context = self._generate_port_context(other_test)
        if nexus_help.is_baremetal(port_context.current):
            connections = self._cisco_mech_driver._get_baremetal_connections(
                port_context.current, False, True)
            for switch_ip, intf_type, port, is_p_vlan, _ in connections:
                port_id = intf_type + ':' + port
                try:
                    host_mapping = nexus_db_v2.get_switch_if_host_mappings(
                        switch_ip, port_id)
                except exceptions.NexusHostMappingNotFound:
                    host_mapping = []
                self.assertEqual(nbr_of_mappings, len(host_mapping))

        # Make sure there is only a single attempt to configure.
        self._verify_results(test_result)

        # Clean all the ncclient mock_calls to clear exception
        # and other mock_call history.
        self.mock_ncclient.reset_mock()

    def _create_delete_port(self, test_name,
                            add_result, del_result):
        """Tests creation and deletion of a virtual port."""

        self._basic_create_verify_port_vlan(
            test_name,
            add_result)

        self._basic_delete_verify_port_vlan(
            test_name,
            del_result)

    def _config_side_effects_on_count(self, match_config, exc,
                                      match_range=None):
        """Generates config-dependent side effect for ncclient.

        This method was written to configure side_effects for both
        ncclient edit_config and get_config drivers.  In the case
        of edit_config, the arguments target and config are passed
        into _side_effect_method.  In the case of get, the argument
        filter is passed into _side_effect_method.  For the sake of
        simplicity, the _side_effect_method was written to handle
        either case.

        Additionally, arguments start and count were passed in to
        handle the number of times to raise exception for a given
        match.  Also match_config if passed in as an empty string,
        is interpreted as match not desired.

        Usage Examples:

        First 2 times for the given mock side-effect, throw an exception.
        _config_side_effects_on_count('', Exception(test_id), range(0,2))

        Two times after 4th attempt for the given mock side-effect,
        throw an exception.
        _config_side_effects_on_count('', Exception(test_id), range(4,6))

        First 2 time, for the given mock side-effect which the call
        matches 'match string', throw an exception.
        _config_side_effects_on_count('match string',
                                      Exception(test_id), range(0,2))

        do 'no range check' and for the given mock side-effect which the call
        matches 'match string', throw an exception.
        _config_side_effects_on_count('match string',
                                      Exception(test_id))
        """
        keywords = match_config.split()

        def _side_effect_method(target=None, config=None, filter=None):
            if not hasattr(self, "position"):
                self.position = 0

            if config is None:
                config = filter[1]
            match = True if not keywords else all(
                word in config for word in keywords)

            # If there is a match, check count in range; otherwise
            # mark as unmatch
            if match and match_range is not None:
                match = self.position in match_range
                self.position += 1

            if match:
                raise exc
            else:
                return mock.DEFAULT

        return _side_effect_method

    def _set_nexus_type_failure(self):
        """Sets exception during ncclient get nexus type. """

        config = {'connect.return_value.get.side_effect':
            self._config_side_effects_on_count('show inventory',
            Exception(__name__))}
        self.mock_ncclient.configure_mock(**config)

    def _create_port_valid_exception(self, attr, match_str, test_case,
                                     which_exc):
        """Verifies exception handling during initial create object.

        This method is a shared method to initiate an exception
        at various point of object creation.  The points of failure
        are identified by the caller which can be get operations or
        edit operations.  When the mechanism replay is functioning,
        the exception should be suppressed and the switch is marked
        as inactive.

        attr:      Which mock attribute to contain side_effect exception
        match_str: String for side_effect method to match for exception
        test_case: which configuration test case to run thru test
        which_exc: Acceptable exception String to raise in exception.
        """

        # Set switch state to active
        switch_ip = self.test_configs[test_case].nexus_ip_addr
        self._cisco_mech_driver.set_switch_ip_and_active_state(
            switch_ip, const.SWITCH_ACTIVE)

        # Clean all the ncclient mock_calls to clear exception
        # and other mock_call history.
        self.mock_ncclient.reset_mock()

        # Set up driver exception
        config = {attr:
            self._config_side_effects_on_count(match_str,
            Exception(which_exc))}
        self.mock_ncclient.configure_mock(**config)

        self._create_port(self.test_configs[test_case])

        # _create_port should complete with no switch state change.
        self.assertEqual(
            const.SWITCH_ACTIVE,
            self._cisco_mech_driver.get_switch_ip_and_active_state(switch_ip))

    def _create_port_failure(self, attr, match_str, test_case, test_id,
                             which_exc=exceptions.NexusConfigFailed):
        """Verifies exception handling during initial create object.

        This method is a shared method to initiate an exception
        at various point of object creation.  The points of failure
        are identified by the caller which can be get operations or
        edit operations.  When the mechanism replay is functioning,
        the exception should be suppressed and the switch is marked
        as inactive.

        attr:      Which mock attribute to contain side_effect exception
        match_str: String for side_effect method to match for exception
        test_case: which configuration test case to run thru test
        test_id:   String to put in the exception.
        which_exc: Exception expecting to encounter
        """

        # Set switch state to active
        switch_ip = self.test_configs[test_case].nexus_ip_addr
        self._cisco_mech_driver.set_switch_ip_and_active_state(
            switch_ip, const.SWITCH_ACTIVE)

        # Clean all the ncclient mock_calls to clear exception
        # and other mock_call history.
        self.mock_ncclient.reset_mock()

        # Set up driver exception
        config = {attr:
            self._config_side_effects_on_count(match_str,
            Exception(test_id))}
        self.mock_ncclient.configure_mock(**config)

        e = self.assertRaises(
                which_exc,
                self._create_port,
                self.test_configs[test_case])
        self.assertIn(test_id, six.u(str(e)))

        # _create_port should complete with no switch state change.
        self.assertEqual(
            self._cisco_mech_driver.get_switch_ip_and_active_state(switch_ip),
            const.SWITCH_ACTIVE)

    def _delete_port_failure(self, attr, match_str, test_case, test_id):
        """Verifies exception handling during object deletion.

        This method is a shared method to initiate an exception
        at various point of object deletion.  The points of failure
        are identified by the caller which can be get operations or
        edit operations.  When the mechanism replay is functioning,
        the exception should be suppressed and the switch is marked
        as inactive.

        attr:      Which mock attribute to contain side_effect exception
        match_str: String for side_effect method to match for exception
        test_case: which configuration test case to run thru test
        test_id:   String to put in the exception.
        """

        # Set switch state to active
        switch_ip = self.test_configs[test_case].nexus_ip_addr
        self._cisco_mech_driver.set_switch_ip_and_active_state(
            switch_ip, const.SWITCH_ACTIVE)

        self._create_port(
            self.test_configs[test_case])

        # _create_port should complete successfully and no switch state change.
        self.assertEqual(
            const.SWITCH_ACTIVE,
            self._cisco_mech_driver.get_switch_ip_and_active_state(switch_ip))

        # Clean all the ncclient mock_calls to clear exception
        # and other mock_call history.
        self.mock_ncclient.reset_mock()

        # Set up driver exception
        config = {attr:
            self._config_side_effects_on_count(match_str,
            Exception(test_id))}
        self.mock_ncclient.configure_mock(**config)

        self.assertRaises(
             exceptions.NexusConfigFailed,
             self._delete_port,
             self.test_configs[test_case])

        # Verify nothing in the port binding data base
        # except Reserved Port Binding if replay is enabled
        nbr_bindings = 1 if self._cisco_mech_driver.is_replay_enabled() else 0
        try:
            port_bindings = nexus_db_v2.get_nexusport_switch_bindings(
                switch_ip)
        except exceptions.NexusPortBindingNotFound:
            port_bindings = []

        self.assertEqual(nbr_bindings, len(port_bindings))

        # Verify nothing in the nve data base
        self.assertEqual(0,
            len(nexus_db_v2.get_nve_switch_bindings(switch_ip)))

        # _delete_port should complete with no switch state change.
        self.assertEqual(
            const.SWITCH_ACTIVE,
            self._cisco_mech_driver.get_switch_ip_and_active_state(switch_ip))


class TestCiscoNexusReplayBase(TestCiscoNexusBase):
    """Replay Base Test Class for Cisco ML2 Nexus driver."""

    def setUp(self):
        """Sets up mock ncclient, and switch and credentials dictionaries."""

        super(TestCiscoNexusReplayBase, self).setUp()

    def _process_replay(self, test1, test2,
                        if_init_result,
                        add_result1, add_result2,
                        replay_result,
                        del_result1, del_result2,
                        replay_init=None,
                        replay_complete=None):
        """Tests create, replay, delete of two ports."""

        # Set all required connection state to True so
        # configurations will succeed
        port_cfg = self.test_configs[test1]
        self._set_switch_state(port_cfg, const.SWITCH_ACTIVE)

        if test2:
            port_cfg = self.test_configs[test2]
            self._set_switch_state(port_cfg, const.SWITCH_ACTIVE)

        self._basic_create_verify_port_vlan(
            test1, add_result1['driver_results'],
            add_result1['nbr_db_entries'])
        if test2:
            self._basic_create_verify_port_vlan(
                test2, add_result2['driver_results'],
                add_result2['nbr_db_entries'])

        # Set all connection state to False for
        # test case HOST_1, NEXUS_IP_ADDRESS_1
        cfg_type = ['test_replay_unique1',
                    'test_replay_duplvlan1',
                    'test_replay_duplport1',
                    'test_replay_unique_vPC']
        for which_cfg in cfg_type:
            if which_cfg in [test1, test2]:
                port_cfg = self.test_configs[which_cfg]
                self._set_switch_state(port_cfg, const.SWITCH_INACTIVE)

        if replay_init:
            replay_init()
        # Since only this test case connection state is False,
        # it should be the only one replayed
        self._cfg_monitor.check_connections()
        if not replay_result:
            replay_result = (if_init_result +
                            add_result1['driver_results'])
            if test2:
                replay_result = (replay_result +
                                 add_result2['driver_results'])
        self._verify_results(replay_result)
        if replay_complete:
            replay_complete()

        # Clear mock_call history so we can evaluate
        # just the result of replay()
        self.mock_ncclient.reset_mock()

        if test2 and 'nbr_db_mappings' in del_result1:
            nbr_db_mappings = del_result1['nbr_db_mappings']
        else:
            nbr_db_mappings = 0
        if test2:
            self._basic_delete_verify_port_vlan(
                test2, del_result1['driver_results'],
                del_result1['nbr_db_entries'],
                nbr_db_mappings)
        self._basic_delete_verify_port_vlan(
                test1, del_result2['driver_results'],
                del_result2['nbr_db_entries'])


class TestContext(TestCiscoNexusBase):
    """Verify Context Blocks for Cisco ML2 Nexus driver."""

    # TODO(caboucha) Put VLAN_TYPE_TRUNK in switch_info for now.
    baremetal_profile = {
        "local_link_information": [
            {
                "switch_id": "10.86.1.129",
                "port_id": "portchannel:1",
                "switch_info": VLAN_TYPE_TRUNK,
            },
            {
                "switch_id": "10.86.1.128",
                "port_id": "portchannel:1",
                "switch_info": VLAN_TYPE_TRUNK,
            },
        ]
    }

    test_configs = {
        'test_vlan_unique1': TestCiscoNexusBase.TestConfigObj(
            NEXUS_IP_ADDRESS_1,
            HOST_NAME_1,
            NEXUS_PORT_1,
            INSTANCE_1,
            VLAN_ID_1,
            NO_VXLAN_ID,
            None,
            DEVICE_OWNER_COMPUTE,
            {},
            None,
            NORMAL_VNIC),
        'test_vxlan_unique1': TestCiscoNexusBase.TestConfigObj(
            NEXUS_IP_ADDRESS_1,
            HOST_NAME_1,
            NEXUS_PORT_1,
            INSTANCE_1,
            VLAN_ID_1,
            VXLAN_ID,
            '225.1.1.1',
            DEVICE_OWNER_COMPUTE,
            {},
            None,
            NORMAL_VNIC),
        'test_bm_vlan_unique1': TestCiscoNexusBase.TestConfigObj(
            NEXUS_IP_ADDRESS_1,
            HOST_NAME_1,
            NEXUS_PORT_1,
            INSTANCE_1,
            VLAN_ID_1,
            NO_VXLAN_ID,
            None,
            DEVICE_OWNER_COMPUTE,
            baremetal_profile,
            None,
            BAREMETAL_VNIC),
    }
    test_configs = collections.OrderedDict(sorted(test_configs.items()))

    def _verify_port_context(self, context, config):
        vlan_segment, vxlan_segment = (
            self._cisco_mech_driver._get_segments(
                context.top_bound_segment,
                context.bottom_bound_segment))
        port = context.current

        if port['device_id'] != config.instance_id:
            return 'device_id mismatch'
        if port[bc.portbindings.HOST_ID] != config.host_name:
            return 'host_name mismatch'
        if vlan_segment[api.SEGMENTATION_ID] != config.vlan_id:
            return 'vlan_id mismatch'
        if vlan_segment[api.NETWORK_TYPE] != p_const.TYPE_VLAN:
            return 'network_type mismatch'
        if port['device_owner'] != config.device_owner:
            return 'device_owner mismatch'
        if vxlan_segment:
            if (vxlan_segment[api.PHYSICAL_NETWORK] !=
                config.mcast_group):
                return 'mcast_group mismatch'
            if (vxlan_segment[api.SEGMENTATION_ID] !=
                config.vxlan_id):
                return 'vxlan_id mismatch'

        if port[bc.portbindings.VNIC_TYPE] != config.vnic_type:
            return 'vnic_type mismatch'

        if config.vnic_type == u'baremetal':
            profile = port[bc.portbindings.PROFILE]['local_link_information']
            cfg_profile = config.profile['local_link_information']
            if (len(profile) != len(cfg_profile) or
                len(profile) == 0):
                return 'profile_len'

            mylen = len(profile)
            for i in range(mylen):
                if (profile[i]['switch_id'] !=
                   cfg_profile[i]['switch_id']):
                    return 'profile.switch_id mismatch'
                if (profile[i]['port_id'] !=
                    cfg_profile[i]['port_id']):
                    return 'profile.port_id mismatch'
                if (profile[i]['switch_info'] !=
                    cfg_profile[i]['switch_info']):
                    return 'profile_switch_info mismatch'
        elif config.vnic_type == u'normal':
            pass
        else:
            return 'vnic_type invalid'

    def test_normal_vlan_format(self):
        port_context = self._generate_port_context(
                self.test_configs['test_vlan_unique1'])

        self.assertIsNone(
            self._verify_port_context(
                port_context,
                self.test_configs['test_vlan_unique1']))

    def test_normal_vxlan_format(self):
        port_context = self._generate_port_context(
                self.test_configs['test_vxlan_unique1'])

        self.assertIsNone(
            self._verify_port_context(
                port_context,
                self.test_configs['test_vxlan_unique1']))

    def test_baremetal_format(self):
        port_context = self._generate_port_context(
                self.test_configs['test_bm_vlan_unique1'])

        self.assertIsNone(
            self._verify_port_context(
                port_context,
                self.test_configs['test_bm_vlan_unique1']))



# Copyright (c) 2013-2016 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_config import types

from networking_cisco._i18n import _
from networking_cisco.config import base

nexus_sub_opts = [
    cfg.BoolOpt('https_verify', default=False,
        help=_('This configuration option defaults to False but '
               'will change to True in Cisco Release 6.0.0. '
               'Set https_verify to True when certification '
               'authority (CA) file is in the Operating Systems '
               'repository or is a locally defined file whose name is '
               'provided in https_local_certificate.  Set https_verify '
               'to False to skip https certification checking thus '
               'making the connection insecure. When True, the Nexus '
               'device must be configured with both certificate and key '
               'files and enabled.  Refer to the "nxapi certificate" '
               'commands defined in the "Nexus 9K NXAPI Programmability '
               'Guide" for details.')),
    cfg.StrOpt('https_local_certificate',
        help=_('Configure a local certificate file to present in https '
               'requests. This is for experimental purposes when an '
               'official certificate from a Trust Certificate Authority '
               'is not yet available. The default configuration is None. '
               'An example configuration would look like '
               'https_local_certificate=/path/to/cafile.crt.')),
    cfg.StrOpt('intfcfg_portchannel', deprecated_name='intfcfg.portchannel',
        help=_('intfcfg_portchannel is a list of Nexus port-channel config '
               'CLI used when baremetal port-channels are created by the '
               'Nexus driver.  It is dependent on "vpc_pool" being '
               'configured.  Any number of Nexus port-channel commands '
               'separated by ";" can be provided.  When there are multiple '
               'interfaces in a baremetal event, the nexus driver checks to '
               'determine whether a port-channel is already applied to the '
               'interfaces; otherwise, it creates a port channel. This '
               'optional configuration allows the administrator to custom '
               'configure the port-channel.  When not configured, the nexus '
               'driver defaults to configuring "spanning-tree port type edge '
               'trunk;no lacp suspend-individual" beneath the port-channel. '
               'An example of this configuration is "intfcfg_portchannel=no '
               'lacp suspend-individual;spanning-tree port type edge '
               'trunk".')),
    cfg.StrOpt('nve_src_intf',
        help=_('Only valid if VXLAN overlay is configured and '
               'vxlan_global_config is set to True. The NVE source interface '
               'is a loopback interface that is configured on the switch with '
               'valid /32 IP address. This /32 IP address must be known by '
               'the transient devices in the transport network and the remote '
               'VTEPs.  This is accomplished by advertising it through a '
               'dynamic routing protocol in the transport network. If '
               'nve_src_intf is not defined, a default setting of 0 '
               'is used to create "loopback0".  This is configured for '
               'non-baremetal only.')),
    cfg.StrOpt('password', secret=True,
        help=_('The password of the Nexus Switch Administrator is required '
               'to allow configuration access to the Nexus switch.')),
    cfg.StrOpt('physnet',
        help=_('This is required if Nexus VXLAN overlay feature is '
               'configured.  It should be the physical network name defined '
               'in "network_vlan_ranges" (defined beneath the "ml2_type_vlan" '
               'section) that this switch is controlling.  The configured '
               '"physnet" is the physical network domain that is connected '
               'to this switch. The vlan ranges defined in '
               '"network_vlan_ranges" for a physical '
               'network are allocated dynamically and are unique per physical '
               'network. These dynamic vlans may be reused across physical '
               'networks.  This configuration applies to non-baremetal '
               'only.')),
    cfg.Opt('host_ports_mapping', default={}, sample_default='<None>',
        type=types.Dict(value_type=types.List(bounds=True)),
        help=_('A list of key:value pairs describing which host is '
               'connected to which physical port or portchannel on the '
               'Nexus switch. The format should look like:\n'
               'host_port_mapping='
               '<your-hostname>:[<intf_type><port>,<intf_type><port>],\n'
               '                  <your-second-host>:[<intf_type><port>]\n'
               'For example:\n'
               'host_port_mapping='
               'host-1:[ethernet1/1, ethernet1/2],\n'
               '                  host-2:[ethernet1/3],\n'
               '                  host-3:[port-channel20]\n'
               'Lines can be broken with indentation to ensure config files '
               'remain readable. '
               'All compute nodes must be configured while '
               'controllers are optional depending on your network '
               'configuration. Depending on the configuration of the '
               'host, the hostname is expected to be the '
               'full hostname (hostname.domainname) which can be derived '
               'by running "hostname -f" on the host itself. Valid '
               'intf_types are "ethernet" or "port-channel".  The default '
               'setting for <intf_type> is "ethernet" and need not be '
               'added to this setting. This configuration applies to VM '
               'deployments only.')),
    cfg.IntOpt('ssh_port', default=22, deprecated_for_removal=True,
        help=_('"ssh_port" specifies the TCP port for connecting via SSH to '
               'manage the switch.  Port number 22 is the default unless the '
               'switch has been configured otherwise. Since "ssh_port" is '
               'associated to the "ncclient" driver which is being '
               'deprecated, "ssh_port" too will be deprecated.')),
    cfg.StrOpt('username',
        help=_('The username of the Nexus Switch Administrator is required '
               'to allow configuration access to the Nexus switch.')),
    cfg.StrOpt('vpc_pool',
        help=_('This is port-channel/VPC allocation pool of ids used with '
               'baremetal deployments only.  When there is a list of ethernet '
               'interfaces provided by Ironic to neutron in a port '
               'event, these are assumed to be a port-channel type '
               'configuration.  Ironic only knows about ethernet interfaces '
               'so it is up to the Nexus Driver to either learn the '
               'port channel if the user preconfigured the channel-group on '
               'the ethernet interfaces; otherwise, the driver will create a '
               'new port-channel and apply the channel-group to the ethernet '
               'interfaces.  This pool is the reserved port-channel IDs '
               'available for allocation by the Nexus driver for each switch. '
               'The full format for "vpc_pool" is '
               'vpc_pool=<start_vpc_no-end_vpc_no> | '
               '<vpc_no> {,<start_vpc_no-end_vpc_no> | <vpc_no>}. The "-" in '
               '<start_vpc_no,end_vpc_no> allows you to configure a range '
               'from start to end and <vpc_no> allows just individual '
               'numbers.  There can be any number of ranges and numbers '
               'separated by commas. There is no default value.  If not '
               'configured, the port-channel will only handle learned cases '
               'and attempts to create port-channels will fail since there is '
               'no id pool available from which to allocate an id. Once '
               'defined, it can be redefined by changing "vpc_pool" and '
               'restarting neutron. Existing VPC ids in the database are '
               'gathered and compared against the new "vpc_pool" config.  New '
               'configured vpcids not found in the database are added.  '
               'Inactive entries in the database not found in the new '
               'configured vpcids list are removed. An example of this '
               'configuration is `vpc_pool=1001-1025,1028`.')),
    base.RemainderOpt('host_port_mapping', deprecated_for_removal=True,
        deprecated_reason="Replaced by 'port_host_mapping' option")]

ml2_cisco_opts = [
    cfg.BoolOpt('host_key_checks', default=False, deprecated_for_removal=True,
        help=_('Set to True to enable strict hostkey checks when connecting '
               'to Nexus switches via ncclient; otherwise, no hostkey checks '
               'are performed). This will be deprecated along with '
               'nexus_driver since this is associated to the ncclient driver '
               'which is going away.')),
    cfg.StrOpt('managed_physical_network',
        help=_('When "managed_physical_network" is configured, it restricts '
               'the network segment that the nexus driver supports. '
               'Setting it to a specific network name will limit the '
               'actions taken by this driver to only that network. The '
               'network name must match a name defined in the '
               '"network_vlan_ranges" configuration.  When '
               '"managed_physical_network" is not set, events for all '
               'network segments will be processed by the driver.')),
    cfg.BoolOpt('never_cache_ssh_connection', default=True,
        deprecated_for_removal=True,
        help=_('Prevent caching ssh connections to a Nexus switch. Set this '
               'to True when there are multiple neutron controllers and/or '
               'when there may be non-neutron ssh connections to the same '
               'Nexus device. Nexus devices have a limit of 8 such '
               'connections. When a single neutron controller has more than '
               '8 processes, caching is automatically disabled without regard '
               'to this configuration. This flag defaults to True which '
               'indicates that ssh connections to a Nexus switch are not '
               'cached when the neutron controller has fewer than 8 '
               'processes.  This will be deprecated along with "nexus_driver" '
               'since this is associated to the ncclient driver which is '
               'going away.')),
    cfg.StrOpt('nexus_driver', default='restapi', deprecated_for_removal=True,
        help=_('The Nexus Mechanism Driver has two driver methods to '
               'configure Nexus devices. The default choice is now "restapi" '
               'which replaces the original "ncclient" driver.  The RESTAPI '
               'driver is preferred because it has better performance with '
               'less Nexus session limits. Additionally, new feature '
               'development is applied only to the restapi driver. To use the '
               'restapi driver, the Nexus 9K image version must be '
               '7.0(3)I5(2) or greater.  For a short term, the original '
               'driver can be used by setting the "nexus_driver" '
               'configuration to "ncclient".  This is short term because '
               'the "ncclient" driver will be removed in Cisco 7.0.0 release '
               'of the networking-cisco repository.')),
    cfg.BoolOpt('persistent_switch_config', default=False,
        deprecated_for_removal=True,
        help=_('To make Nexus device persistent by running the Nexus '
               'CLI "copy run start" after applying successful '
               'configurations. This will be deprecated along with '
               'nexus_driver since this is associated to the ncclient '
               'driver which is going away.')),
    cfg.BoolOpt('provider_vlan_auto_create', default=True,
        help=_('A flag indicating whether the Nexus driver should manage '
               'the creation and removal of VLANs for provider networks on '
               'the Nexus switches. When this flag is False, the Nexus '
               'driver will not create or remove VLANs for provider '
               'networks and the administrator needs to manage these '
               'interfaces manually or by external orchestration.')),
    cfg.BoolOpt('provider_vlan_auto_trunk', default=True,
        help=_('A flag indicating whether Nexus driver should manage '
               'the adding and removing of provider VLANs from trunk ports on '
               'the Nexus switches. When this flag is False, the Nexus '
               'driver will not add or remove provider VLANs from trunk '
               'ports and the administrator needs to manage these operations '
               'manually or by external orchestration.')),
    cfg.IntOpt('switch_heartbeat_time', default=30,
        help=_('Configuration replay is enabled by default by defining the '
               'time interval to 30 seconds.  This is the amount of time to '
               'check the state of all known Nexus device(s). To disable '
               'the replay feature, set this "switch_heartbeat_time" to 0 '
               'seconds.')),
    cfg.BoolOpt('vxlan_global_config', default=False,
        help=_('A flag indicating whether the Nexus driver should manage '
               'the creating and removing of the Nexus switch VXLAN global '
               'settings of "feature nv overlay", "feature '
               'vn-segment-vlan-based", "interface nve 1" and the NVE '
               'subcommand "source-interface loopback #". When set to the '
               'default of False, the Nexus driver will not add or remove '
               'these VXLAN settings and the administrator needs to manage '
               'these operations manually or by external orchestration.')),
]

nexus_switches = base.SubsectionOpt(
    'ml2_mech_cisco_nexus',
    dest='nexus_switches',
    help=_("Subgroups that allow you to specify the nexus switches to be "
           "managed by the nexus ML2 driver."),
    subopts=nexus_sub_opts)

cfg.CONF.register_opts(ml2_cisco_opts, "ml2_cisco")
cfg.CONF.register_opt(nexus_switches, "ml2_cisco")

#
# Format for ml2_conf_cisco.ini 'ml2_mech_cisco_nexus' is:
# {('<device ipaddr>', '<keyword>'): '<value>', ...}
#
# Example:
# {('1.1.1.1', 'username'): 'admin',
#  ('1.1.1.1', 'password'): 'mySecretPassword',
#  ('1.1.1.1', 'compute1'): '1/1', ...}
#



# Copyright 2016 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from binascii import hexlify
import mock
import socket

from networking_cisco.plugins.cisco.cpnr.cpnr_client import UnexpectedError
from networking_cisco.plugins.cisco.cpnr.cpnr_dhcp_relay_agent import (
    DhcpPacket)
from networking_cisco.plugins.cisco.cpnr.cpnr_dhcp_relay_agent import (
    DhcpRelayAgent)
from networking_cisco.plugins.cisco.cpnr.cpnr_dhcp_relay_agent import cfg
from networking_cisco.plugins.cisco.cpnr.cpnr_dhcp_relay_agent import OPTS
import unittest


class TestDhcpRelayAgent(unittest.TestCase):

    @mock.patch('networking_cisco.plugins.cisco.'
                'cpnr.cpnr_dhcp_relay_agent.netns')
    @mock.patch('socket.socket')
    def test_open_dhcp_ext_socket(self, mock_socket, mock_netns):
        cfg.CONF.register_opts(OPTS, 'cisco_pnr')
        relay = DhcpRelayAgent()
        mock_netns.iflist.return_value = []
        mock_netns.iflist.return_value.append(('lo', '127.0.0.1',
                                               '255.0.0.0'))
        sock, addr = relay._open_dhcp_ext_socket()

        self.assertTrue(mock_netns.iflist.called, "Failed to call iflist.")

        mock_socket.assert_has_calls([
            mock.call(socket.AF_INET, socket.SOCK_DGRAM),
            mock.call().bind(('127.0.0.1', 67)),
            mock.call().connect(('127.0.0.1', 67))]
        )

        # check exception thrown if no interfaces
        with self.assertRaises(UnexpectedError):
            mock_netns.iflist.return_value = []
            sock, addr = relay._open_dhcp_ext_socket()

        # check exception thrown if no matching interfaces
        with self.assertRaises(UnexpectedError):
            mock_netns.iflist.return_value = []
            mock_netns.iflist.return_value.append(('eth1', '10.0.1.3',
                                                   '255.255.255.0'))
            sock, addr = relay._open_dhcp_ext_socket()

        # check matching interface found if not first in list
        mock_netns.iflist.return_value.append(('eth0', '10.0.0.10',
                                               '255.255.255.0'))
        mock_netns.iflist.return_value.append(('lo', '127.0.0.1',
                                               '255.0.0.0'))
        sock, addr = relay._open_dhcp_ext_socket()

    @mock.patch('networking_cisco.plugins.cisco.'
                'cpnr.cpnr_dhcp_relay_agent.netns')
    @mock.patch('socket.socket')
    def test_open_dhcp_int_socket(self, mock_socket, mock_netns):
        cfg.CONF.register_opts(OPTS, 'cisco_pnr')
        relay = DhcpRelayAgent()

        mock_netns.iflist.return_value = []
        mock_netns.iflist.return_value.append(('eth0', '10.1.1.7',
                                               '255.255.255.0'))
        recv_s, send_s, addr = relay._open_dhcp_int_socket()

        self.assertTrue(mock_netns.iflist.called, "Failed to call iflist.")

        mock_socket.assert_has_calls([
            mock.call(socket.AF_INET, socket.SOCK_DGRAM),
            mock.call().setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1),
            mock.call().setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1),
            mock.call().bind(('0.0.0.0', 67)),
            mock.call(socket.AF_INET, socket.SOCK_DGRAM),
            mock.call().setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1),
            mock.call().setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1),
            mock.call().bind(('10.1.1.7', 67))]
        )

        # check exception thrown if no interfaces
        with self.assertRaises(UnexpectedError):
            mock_netns.iflist.return_value = []
            recv_s, send_s, addr = relay._open_dhcp_int_socket()

    def test_convert_ns_to_vpnid(self):
        cfg.CONF.register_opts(OPTS, 'cisco_pnr')
        relay = DhcpRelayAgent()
        namespace = 'qdhcp-a207e329-9476-4746-91a7-fb1cce171a81'
        vpnid = relay._convert_ns_to_vpnid(namespace)
        expected = 'a7fb1cce171a81'
        self.assertEqual(expected, vpnid)


class TestDhcpPacket(unittest.TestCase):

    def test_parse(self):
        # DHCP packet contains relay agent option 82
        data_str = ('0101 0600 6f75 7345 0000 0000 0000 0000'
            '0000 0000 0000 0000 c0a8 3204 fa16 3ea5'
            '9fa4 0000 0000 0000 0000 0000 0000 0000'
            '0000 0000 0000 0000 0000 0000 0000 0000'
            '0000 0000 0000 0000 0000 0000 0000 0000'
            '0000 0000 0000 0000 0000 0000 0000 0000'
            '0000 0000 0000 0000 0000 0000 0000 0000'
            '0000 0000 0000 0000 0000 0000 0000 0000'
            '0000 0000 0000 0000 0000 0000 0000 0000'
            '0000 0000 0000 0000 0000 0000 0000 0000'
            '0000 0000 0000 0000 0000 0000 0000 0000'
            '0000 0000 0000 0000 0000 0000 0000 0000'
            '0000 0000 0000 0000 0000 0000 0000 0000'
            '0000 0000 0000 0000 0000 0000 0000 0000'
            '0000 0000 0000 0000 0000 0000 6382 5363'
            '3501 013d 0701 fa16 3ea5 9fa4 3902 0240'
            '3707 0103 060c 0f1c 2a3c 0c75 6468 6370'
            '2031 2e32 302e 3152 1898 000b 040a 0a01'
            '0205 040a 0a01 0297 0801 a7fb 1cce 171a'
            '81ff')

        buf = bytearray.fromhex(data_str)
        packet = DhcpPacket.parse(buf)
        # Test client address
        self.assertEqual('0.0.0.0', packet.get_ciaddr())
        # Test relay agent options
        expected_relay_options = {152: b'',
                                  11: '10.10.1.2',
                                  5: '10.10.1.2',
                                  151: b'a7fb1cce171a81'}
        actual_packet_options = {code: packet.get_relay_option(code)
                                 for code in [152, 11, 5, 151]}
        self.assertEqual(expected_relay_options, actual_packet_options)

        # Unsuccessful case of undefined relay agent sub-options
        with self.assertRaises(KeyError):
            packet.get_relay_option(220)

    def test_data(self):
        # DHCP packet contains relay agent option 82
        data_str = ('0101 0600 6f75 7345 0000 0000 0000 0000'
            '0000 0000 0000 0000 c0a8 3204 fa16 3ea5'
            '9fa4 0000 0000 0000 0000 0000 0000 0000'
            '0000 0000 0000 0000 0000 0000 0000 0000'
            '0000 0000 0000 0000 0000 0000 0000 0000'
            '0000 0000 0000 0000 0000 0000 0000 0000'
            '0000 0000 0000 0000 0000 0000 0000 0000'
            '0000 0000 0000 0000 0000 0000 0000 0000'
            '0000 0000 0000 0000 0000 0000 0000 0000'
            '0000 0000 0000 0000 0000 0000 0000 0000'
            '0000 0000 0000 0000 0000 0000 0000 0000'
            '0000 0000 0000 0000 0000 0000 0000 0000'
            '0000 0000 0000 0000 0000 0000 0000 0000'
            '0000 0000 0000 0000 0000 0000 0000 0000'
            '0000 0000 0000 0000 0000 0000 6382 5363'
            '3501 013d 0701 fa16 3ea5 9fa4 3902 0240'
            '3707 0103 060c 0f1c 2a3c 0c75 6468 6370'
            '2031 2e32 302e 3152 1898 000b 040a 0a01'
            '0205 040a 0a01 0297 0801 a7fb 1cce 171a'
            '81ff')
        buf = bytearray.fromhex(data_str)
        pktbuf = bytearray(4096)
        pktbuf[0:len(buf)] = buf
        packet = DhcpPacket.parse(pktbuf)
        hex_data = hexlify(packet.data())
        self.assertIn(hexlify(packet.ciaddr), hex_data)
        self.assertIn(hexlify(packet.giaddr), hex_data)

        expected_relay_options = {152: b'',
                                  11: '10.10.1.2',
                                  5: '10.10.1.2',
                                  151: b'a7fb1cce171a81'}
        # Find relay agent sub-options in data
        self.assertIn(self.get_relay_opt_hex(expected_relay_options[11]),
            hex_data)
        self.assertIn(self.get_relay_opt_hex(expected_relay_options[5]),
            hex_data)
        self.assertIn(b"01" + expected_relay_options[151],
            hex_data)
        self.assertIn(expected_relay_options[152],
            hex_data)

    def get_relay_opt_hex(self, value):
        return hexlify(socket.inet_aton(value))



# Copyright 2017 Cisco Systems.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

import mock

from neutron.tests import base

from networking_cisco.apps.saf.agent.topo_disc import (
    topo_disc_constants as topo_constants)
from networking_cisco.apps.saf.agent.topo_disc import topo_disc

try:
    OrderedDict = collections.OrderedDict
except AttributeError:
    import ordereddict
    OrderedDict = ordereddict.OrderedDict


class TopoDiscTest(base.BaseTestCase):
    """A test suite to exercise the TopoDisc Class. """

    def setUp(self):
        """Setup Routine. """
        super(TopoDiscTest, self).setUp()
        self.root_helper = 'sudo'
        self.intf_list = ['eth1', 'eth2']
        self.updated_intf_list = ['eth1', 'eth2', 'LLDPLeth2']
        self.protocol_interface = 'LLDPLeth2'
        self.phy_interface = 'eth2'
        period_fn = mock.patch(
            'networking_cisco.apps.saf.common.utils.PeriodicTask').start()
        self.get_run_phy_intf = mock.patch(
            'networking_cisco.apps.saf.common.dfa_sys_lib.'
            'get_all_run_phy_intf', return_value=self.intf_list).start()
        self.enable_lldp = mock.patch(
            'networking_cisco.apps.saf.agent.topo_disc.'
            'pub_lldp_api.LldpApi.enable_lldp').start()
        period_obj = period_fn.return_value
        parent = mock.MagicMock()
        parent.attach_mock(period_obj.run, 'run')
        expected_calls = [mock.call.run()]
        self.topo_disc = topo_disc.TopoDisc(None, self.root_helper)
        period_fn.assert_any_call(topo_constants.PERIODIC_TASK_INTERVAL,
                                  self.topo_disc.periodic_discovery_task)
        parent.assert_has_calls(expected_calls)

    def test_init_cfg_interfaces(self):
        """Test the _init_cfg_interfaces function. """
        with mock.patch('networking_cisco.apps.saf.agent.topo_disc.'
                        'topo_disc.TopoDisc.'
                        'cfg_lldp_interface_list') as topo_intf_fn:
            self.topo_disc._init_cfg_interfaces("")
        topo_intf_fn.assert_any_call(self.intf_list)
        self.assertEqual(self.topo_disc.intf_list, self.intf_list)

    def test_cfg_intf(self):
        """Test the cfg_intf function. """
        with mock.patch.object(self.topo_disc,
                               'cfg_lldp_interface') as cfg_lldp_intf_fn:
            self.topo_disc.cfg_intf(self.protocol_interface,
                                    self.phy_interface)
        cfg_lldp_intf_fn.assert_any_call(self.protocol_interface,
                                         self.phy_interface)
        self.assertEqual(self.updated_intf_list, self.topo_disc.intf_list)

    def test_cfg_lldp_interface_list(self):
        """Test the cfg_lldp_interface_list function. """
        with mock.patch.object(self.topo_disc,
                               'cfg_lldp_interface') as cfg_lldp_intf_fn:
            self.topo_disc.cfg_lldp_interface_list(self.intf_list)
        for intf in self.intf_list:
            cfg_lldp_intf_fn.assert_any_call(intf)

    def test_cfg_lldp_interface(self):
        """Test the test_cfg_lldp_interface function. """
        self.enable_lldp.return_value = True
        self.topo_disc.cfg_lldp_interface(self.protocol_interface,
                                          self.phy_interface)
        self.enable_lldp.assert_called_with(self.protocol_interface)
        self.assertTrue(
            self.topo_disc.get_attr_obj(
                self.protocol_interface).get_lldp_status())

    def test_cfg_lldp_interface_error(self):
        """Test the test_cfg_lldp_interface function when it returns False. """
        self.enable_lldp.return_value = False
        self.topo_disc.cfg_lldp_interface(self.protocol_interface,
                                          self.phy_interface)
        self.enable_lldp.assert_called_with(self.protocol_interface)
        self.assertFalse(
            self.topo_disc.get_attr_obj(
                self.protocol_interface).get_lldp_status())

    def _set_tlv_params_attr_obj(self, attr_obj, val):
        attr_obj.return_value.remote_evb_mode_uneq_store.return_value = val
        attr_obj.return_value.remote_evb_cfgd_uneq_store.return_value = val
        attr_obj.return_value.remote_mgmt_addr_uneq_store.return_value = val
        attr_obj.return_value.remote_sys_desc_uneq_store.return_value = val
        attr_obj.return_value.remote_sys_name_uneq_store.return_value = val
        attr_obj.return_value.remote_port_uneq_store.return_value = val
        attr_obj.return_value.remote_chassis_id_mac_uneq_store.return_value = (
            val)
        attr_obj.return_value.remote_port_id_mac_uneq_store.return_value = val

    def test_cmp_store_tlv_params_all_false(self):
        """Test the test_cmp_store_tlv_params for all false case.

        This test the case when all TLV compare functions returns False.
        """
        with mock.patch.object(self.topo_disc, 'get_attr_obj') as attr_obj,\
            mock.patch('networking_cisco.apps.saf.agent.topo_disc.'
                       'pub_lldp_api.LldpApi.get_remote_evb_mode'):
            self._set_tlv_params_attr_obj(attr_obj, False)
            ret = self.topo_disc.cmp_store_tlv_params(self.protocol_interface,
                                                      'Some TLV Data')
        self.assertFalse(ret)

    def test_cmp_store_tlv_params_all_true(self):
        """Test the test_cmp_store_tlv_params for all True case.

        This test the case when all TLV compare functions returns True.
        """
        with mock.patch.object(self.topo_disc, 'get_attr_obj') as attr_obj,\
            mock.patch('networking_cisco.apps.saf.agent.topo_disc.'
                       'pub_lldp_api.LldpApi.get_remote_evb_mode'):
            self._set_tlv_params_attr_obj(attr_obj, True)
            ret = self.topo_disc.cmp_store_tlv_params(self.protocol_interface,
                                                      'Some TLV Data')
        self.assertTrue(ret)

    def test_cmp_store_tlv_params_one_true(self):
        """Test the test_cmp_store_tlv_params for one True case.

        This test the case when all TLV compare functions returns True.
        """
        with mock.patch.object(self.topo_disc, 'get_attr_obj') as attr_obj,\
            mock.patch('networking_cisco.apps.saf.agent.topo_disc.'
                       'pub_lldp_api.LldpApi.get_remote_evb_mode'):
            self._set_tlv_params_attr_obj(attr_obj, False)
            attr_obj.return_value.remote_port_uneq_store.return_value = True
            ret = self.topo_disc.cmp_store_tlv_params(self.protocol_interface,
                                                      'Some TLV Data')
        self.assertTrue(ret)

    def _test_uneq_store_fn(self, fn_ext, new_value, stored_value,
                            member_name, fn_ret):
        """Generic function to test all variants of uneq_store functions. """
        attr_obj = self.topo_disc.get_attr_obj(self.phy_interface)
        fn_name = getattr(attr_obj, fn_ext)
        setattr(attr_obj, member_name, stored_value)
        ret = fn_name(new_value)
        self.assertEqual(fn_ret, ret)
        self.assertEqual(new_value, getattr(attr_obj, member_name))

    def test_remote_evb_mode_uneq_store_equal(self):
        """Test remote_evb_mode_uneq_store for same value case.

        This tests the case when the value stored is the same as the value
        passed.
        """
        evb_mode = 'bridge'
        self._test_uneq_store_fn(
            'remote_evb_mode_uneq_store', evb_mode, evb_mode,
            'remote_evb_mode', False)

    def test_remote_evb_mode_uneq_store_unequal(self):
        """Test remote_evb_mode_uneq_store for same value case.

        This tests the case when the value stored is the same as the value
        passed.
        """
        evb_mode = 'bridge'
        old_evb_mode = ''
        self._test_uneq_store_fn(
            'remote_evb_mode_uneq_store', evb_mode, old_evb_mode,
            'remote_evb_mode', True)

    def test_remote_evb_cfgd_uneq_store_equal(self):
        """Test remote_evb_cfgd_uneq_store for same value case.

        This tests the case when the value stored is the same as the value
        passed.
        """
        evb_cfgd = True
        self._test_uneq_store_fn(
            'remote_evb_cfgd_uneq_store', evb_cfgd, evb_cfgd,
            'remote_evb_cfgd', False)

    def test_remote_evb_cfgd_uneq_store_unequal(self):
        """Test remote_evb_cfgd_uneq_store for same value case.

        This tests the case when the value stored is the same as the value
        passed.
        """
        evb_cfgd = True
        old_evb_cfgd = False
        self._test_uneq_store_fn(
            'remote_evb_cfgd_uneq_store', evb_cfgd, old_evb_cfgd,
            'remote_evb_cfgd', True)

    def test_remote_mgmt_addr_uneq_store_equal(self):
        """Test remote_mgmt_addr_uneq_store for same value case.

        This tests the case when the value stored is the same as the value
        passed.
        """
        mgmt_addr = '10.2.2.4'
        self._test_uneq_store_fn(
            'remote_mgmt_addr_uneq_store', mgmt_addr, mgmt_addr,
            'remote_mgmt_addr', False)

    def test_remote_mgmt_addr_uneq_store_unequal(self):
        """Test remote_mgmt_addr_uneq_store for same value case.

        This tests the case when the value stored is the same as the value
        passed.
        """
        mgmt_addr = '10.2.2.4'
        old_mgmt_addr = '1.2.5.4'
        self._test_uneq_store_fn(
            'remote_mgmt_addr_uneq_store', mgmt_addr, old_mgmt_addr,
            'remote_mgmt_addr', True)

    def test_remote_system_desc_uneq_store_equal(self):
        """Test remote_system_desc_uneq_store for same value case.

        This tests the case when the value stored is the same as the value
        passed.
        """
        system_desc = "Cisco NXOS"
        self._test_uneq_store_fn(
            'remote_sys_desc_uneq_store', system_desc, system_desc,
            'remote_system_desc', False)

    def test_remote_system_desc_uneq_store_unequal(self):
        """Test remote_system_desc_uneq_store for same value case.

        This tests the case when the value stored is the same as the value
        passed.
        """
        system_desc = "Cisco NXOS"
        old_system_desc = "Cisco Non-NXOS"
        self._test_uneq_store_fn(
            'remote_sys_desc_uneq_store', system_desc, old_system_desc,
            'remote_system_desc', True)

    def test_remote_system_name_uneq_store_equal(self):
        """Test remote_system_name_uneq_store for same value case.

        This tests the case when the value stored is the same as the value
        passed.
        """
        system_name = 'N6K-50'
        self._test_uneq_store_fn(
            'remote_sys_name_uneq_store', system_name, system_name,
            'remote_system_name', False)

    def test_remote_system_name_uneq_store_unequal(self):
        """Test remote_system_name_uneq_store for same value case.

        This tests the case when the value stored is the same as the value
        passed.
        """
        system_name = 'N6K-50'
        old_system_name = 'N6K-51'
        self._test_uneq_store_fn(
            'remote_sys_name_uneq_store', system_name, old_system_name,
            'remote_system_name', True)

    def test_remote_port_uneq_store_equal(self):
        """Test remote_port_uneq_store for same value case.

        This tests the case when the value stored is the same as the value
        passed.
        """
        port = 'G1/1'
        self._test_uneq_store_fn(
            'remote_port_uneq_store', port, port, 'remote_port', False)

    def test_remote_port_uneq_store_unequal(self):
        """Test remote_port_uneq_store for same value case.

        This tests the case when the value stored is the same as the value
        passed.
        """
        port = 'G1/1'
        old_port = 'G1/2'
        self._test_uneq_store_fn(
            'remote_port_uneq_store', port, old_port, 'remote_port', True)

    def test_remote_chassis_id_mac_uneq_store_equal(self):
        """Test remote_chassis_id_mac_uneq_store for same value case.

        This tests the case when the value stored is the same as the value
        passed.
        """
        chassis_id_mac = '00:11:22:33:44:55'
        self._test_uneq_store_fn(
            'remote_chassis_id_mac_uneq_store', chassis_id_mac, chassis_id_mac,
            'remote_chassis_id_mac', False)

    def test_remote_chassis_id_mac_uneq_store_unequal(self):
        """Test remote_chassis_id_mac_uneq_store for same value case.

        This tests the case when the value stored is the same as the value
        passed.
        """
        chassis_id_mac = '00:11:22:33:44:55'
        old_chassis_id_mac = '00:22:33:44:55:66'
        self._test_uneq_store_fn(
            'remote_chassis_id_mac_uneq_store', chassis_id_mac,
            old_chassis_id_mac, 'remote_chassis_id_mac', True)

    def test_remote_port_id_mac_uneq_store_equal(self):
        """Test remote_port_id_mac_uneq_store for same value case.

        This tests the case when the value stored is the same as the value
        passed.
        """
        port_id_mac = '00:aa:bb:cc:dd:ee'
        self._test_uneq_store_fn(
            'remote_port_id_mac_uneq_store', port_id_mac, port_id_mac,
            'remote_port_id_mac', False)

    def test_remote_port_id_mac_uneq_store_unequal(self):
        """Test remote_port_id_mac_uneq_store for same value case.

        This tests the case when the value stored is the same as the value
        passed.
        """
        port_id_mac = '00:aa:bb:cc:dd:ee'
        old_port_id_mac = '00:bb:cc:dd:ee:ff'
        self._test_uneq_store_fn(
            'remote_port_id_mac_uneq_store', port_id_mac, old_port_id_mac,
            'remote_port_id_mac', True)

    def _test_period_task_reset_general(self, cb_ret,
                                        bond_interface_change_ret=False,
                                        cmp_store_tlv_ret=False,
                                        get_db_retry_status=False,
                                        topo_disc_send_cnt=1,
                                        update_lldp_status=True):
        """Internal helper function for all period_tast_reset functions. """
        with mock.patch('networking_cisco.apps.saf.agent.topo_disc.'
                        'pub_lldp_api.LldpApi.get_lldp_tlv'),\
            mock.patch.object(
                self.topo_disc, '_check_bond_interface_change',
                return_value=bond_interface_change_ret),\
            mock.patch.object(self.topo_disc, 'cmp_store_tlv_params',
                              return_value=cmp_store_tlv_ret):
            self.topo_disc.cb = mock.MagicMock()
            self.topo_disc.cb.return_value = cb_ret
            self.topo_disc.intf_list = [self.phy_interface]
            attr_obj = self.topo_disc.get_attr_obj(self.phy_interface)
            attr_obj.update_lldp_status(update_lldp_status)
            attr_obj.store_db_retry_status(get_db_retry_status)
            attr_obj.topo_send_cnt = topo_disc_send_cnt
            self.topo_disc.periodic_discovery_task()
        self.assertEqual(0, attr_obj.topo_send_cnt)
        self.assertEqual(not cb_ret, attr_obj.get_db_retry_status())
        self.topo_disc.cb.assert_called_with(
            self.phy_interface,
            self.topo_disc.get_attr_obj(self.phy_interface))

    def test_period_task_reset_case_cmp_true(self):
        """Test the periodic task when cmp_store_tlv_params returns True.

        Also, the callback invoked case is verified for a return value of True.
        """
        cb_ret = True
        self._test_period_task_reset_general(cb_ret, cmp_store_tlv_ret=True)

    def test_period_task_reset_case_get_db_retry_true(self):
        """Test the periodic task when get_db_retry_status returns True.

        Also, the callback invoked case is verified for a return value of
        False.
        """
        cb_ret = False
        self._test_period_task_reset_general(cb_ret, get_db_retry_status=True)

    def test_period_task_reset_case_bond_intf_change_true(self):
        """Test the periodic task when bond_intf_change returns True.

        Also, the callback invoked case is verified for a return value of
        False.
        """
        cb_ret = False
        self._test_period_task_reset_general(
            cb_ret, bond_interface_change_ret=True)

    def test_period_task_reset_case_topo_disc_cnt_exceed_threshold(self):
        """Test the periodic task when send failure cnt exceeds threshold.

        Also, the callback invoked case is verified for a return value of
        True.
        """
        cb_ret = True
        self._test_period_task_reset_general(
            cb_ret,
            topo_disc_send_cnt=topo_constants.TOPO_DISC_SEND_THRESHOLD + 1)

    def test_period_task_reset_case_all_false(self):
        """Test the periodic task when callback is not invoked.

        When all conditions for calling callback returns False.
        """
        topo_send_cnt = 1
        with mock.patch('networking_cisco.apps.saf.agent.topo_disc.'
                        'pub_lldp_api.LldpApi.get_lldp_tlv'),\
            mock.patch.object(
                self.topo_disc, '_check_bond_interface_change',
                return_value=False),\
            mock.patch.object(self.topo_disc, 'cmp_store_tlv_params',
                              return_value=False):
            self.topo_disc.cb = mock.MagicMock()
            self.topo_disc.cb.return_value = False
            self.topo_disc.intf_list = [self.phy_interface]
            attr_obj = self.topo_disc.get_attr_obj(self.phy_interface)
            attr_obj.update_lldp_status(True)
            attr_obj.store_db_retry_status(False)
            attr_obj.topo_send_cnt = topo_send_cnt
            self.topo_disc.periodic_discovery_task()
        self.assertEqual(topo_send_cnt + 1, attr_obj.topo_send_cnt)
        self.topo_disc.cb.assert_not_called()

    def test_period_task_reset_case_lldp_status_false(self):
        """Test the periodic task when LLDP status is False. """
        topo_send_cnt = 1
        with mock.patch('networking_cisco.apps.saf.agent.topo_disc.'
                        'pub_lldp_api.LldpApi.get_lldp_tlv'),\
            mock.patch.object(
                self.topo_disc, '_check_bond_interface_change',
                return_value=False),\
            mock.patch.object(self.topo_disc, 'cmp_store_tlv_params',
                              return_value=False):
            self.topo_disc.cb = mock.MagicMock()
            self.topo_disc.cb.return_value = False
            self.topo_disc.intf_list = [self.phy_interface]
            attr_obj = self.topo_disc.get_attr_obj(self.phy_interface)
            attr_obj.update_lldp_status(False)
            attr_obj.store_db_retry_status(False)
            attr_obj.topo_send_cnt = topo_send_cnt
            self.enable_lldp.return_value = True
            self.topo_disc.periodic_discovery_task()
        self.assertEqual(topo_send_cnt, attr_obj.topo_send_cnt)
        self.topo_disc.cb.assert_not_called()
        self.enable_lldp.assert_called_with(self.phy_interface)
        self.assertTrue(
            self.topo_disc.get_attr_obj(self.phy_interface).get_lldp_status())



# Copyright (c) 2011-2016 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


RESERVED_PORT_HOST_ID = 'reserved_port'
CREDENTIAL_USERNAME = 'user_name'
CREDENTIAL_PASSWORD = 'password'

VPCPOOL = 'vpc_pool'
USERNAME = 'username'
PASSWORD = 'password'
SSHPORT = 'ssh_port'
PHYSNET = 'physnet'
IF_PC_DEPRECATE = 'intfcfg.portchannel'
IF_PC = 'intfcfg_portchannel'
HTTPS_VERIFY = 'https_verify'
HTTPS_CERT = 'https_local_certificate'
NVE_SRC_INTF = 'nve_src_intf'

UNAME_TUPLE = 0
PW_TUPLE = 1
HTTPS_VERIFY_TUPLE = 2
HTTPS_CERT_TUPLE = 3
COOKIE_TUPLE = 4

NETWORK_ADMIN = 'network_admin'

CISCO_NEXUS_ML2_MECH_DRIVER_V2 = 'cisco_nexus'
TYPE_NEXUS_VXLAN = 'nexus_vxlan'

NVE_INT_NUM = '1'
NEXUS_MAX_VLAN_NAME_LEN = 32

NO_DUPLICATE = 0
DUPLICATE_VLAN = 1
DUPLICATE_PORT = 2

NEXUS_TYPE_INVALID = -1
NEXUS_3K = 3
NEXUS_5K = 5
NEXUS_7K = 7
NEXUS_9K = 9

MAX_NEXUS_SSH_SESSIONS = 8

REPLAY_FAILURES = '_replay_failures'
FAIL_CONTACT = '_contact'
FAIL_CONFIG = '_config'

RESERVED_NEXUS_SWITCH_DEVICE_ID_R1 = "RESERVED_NEXUS_SWITCH_DEVICE_ID_R1"
NO_PORT_ID = "NONE"
NO_VLAN_OR_VNI_ID = '0'
SWITCH_ACTIVE = "ACTIVE"
SWITCH_RESTORE_S1 = "RESTORE_S1"
SWITCH_RESTORE_S2 = "RESTORE_S2"
SWITCH_INACTIVE = "INACTIVE"

CREATE_VLAN_SEND_SIZE = 20
CREATE_VLAN_BATCH = 200
CREATE_PORT_VLAN_LENGTH = 20

NOT_NATIVE = False

UNCONFIGURED_VLAN = "1-4094"
MINVPC = 1
MAXVPC = 4096



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import eventlet

from oslo_log import log as logging
from sqlalchemy.sql import expression as expr

from neutron.db import models_v2
from neutron_lib import exceptions as n_exc

from networking_cisco._i18n import _LE, _LI, _LW
from networking_cisco import backwards_compatibility as bc
from networking_cisco.plugins.cisco.device_manager.plugging_drivers import (
    noop_plugging_driver)
LOG = logging.getLogger(__name__)


DELETION_ATTEMPTS = 5
SECONDS_BETWEEN_DELETION_ATTEMPTS = 3


class TestPluggingDriver(noop_plugging_driver.NoopPluggingDriver):
    """Driver class for unit tests."""

    def create_hosting_device_resources(self, context, complementary_id,
                                        tenant_id, mgmt_context, max_hosted):
        mgmt_port = None
        if mgmt_context and mgmt_context.get('mgmt_nw_id') and tenant_id:
            # Create port for mgmt interface
            p_spec = {'port': {
                'tenant_id': tenant_id,
                'admin_state_up': True,
                'name': 'mgmt',
                'network_id': mgmt_context['mgmt_nw_id'],
                'mac_address': bc.constants.ATTR_NOT_SPECIFIED,
                'fixed_ips': self._mgmt_subnet_spec(context, mgmt_context),
                'device_id': "",
                # Use device_owner attribute to ensure we can query for these
                # ports even before Nova has set device_id attribute.
                'device_owner': complementary_id}}
            try:
                mgmt_port = self._core_plugin.create_port(context,
                                                          p_spec)
            except n_exc.NeutronException as e:
                LOG.error(_LE('Error %s when creating service VM resources. '
                              'Cleaning up.'), e)
                resources = {}
                self.delete_hosting_device_resources(
                    context, tenant_id, mgmt_port, **resources)
                mgmt_port = None
        return {'mgmt_port': mgmt_port}

    def get_hosting_device_resources(self, context, id, complementary_id,
                                     tenant_id, mgmt_nw_id):
        ports, nets, subnets = [], [], []
        mgmt_port = None
        # Ports for hosting device may not yet have 'device_id' set to
        # Nova assigned uuid of VM instance. However, those ports will still
        # have 'device_owner' attribute set to complementary_id. Hence, we
        # use both attributes in the query to ensure we find all ports.
        query = context.session.query(models_v2.Port)
        query = query.filter(expr.or_(
            models_v2.Port.device_id == id,
            models_v2.Port.device_owner == complementary_id))
        for port in query:
            if port['network_id'] != mgmt_nw_id:
                ports.append(port)
                nets.append({'id': port['network_id']})
                subnets.append({'id': port['fixed_ips'][0]['subnet_id']})
            else:
                mgmt_port = port
        return {'mgmt_port': mgmt_port,
                'ports': ports, 'networks': nets, 'subnets': subnets}

    def delete_hosting_device_resources(self, context, tenant_id, mgmt_port,
                                        **kwargs):
        attempts = 1
        while mgmt_port is not None:
            if attempts == DELETION_ATTEMPTS:
                LOG.warning(_LW('Aborting resource deletion after %d '
                                'unsuccessful attempts'), DELETION_ATTEMPTS)
                return
            else:
                if attempts > 1:
                    eventlet.sleep(SECONDS_BETWEEN_DELETION_ATTEMPTS)
                LOG.info(_LI('Resource deletion attempt %d starting'),
                         attempts)
            # Remove anything created.
            if mgmt_port is not None:
                ml = {mgmt_port['id']}
                self._delete_resources(context, "management port",
                                       self._core_plugin.delete_port,
                                       n_exc.PortNotFound, ml)
                if not ml:
                    mgmt_port = None
            attempts += 1
        LOG.info(_LI('Resource deletion succeeded'))

    def _delete_resources(self, context, name, deleter, exception_type,
                          resource_ids):
        for item_id in resource_ids.copy():
            try:
                deleter(context, item_id)
                resource_ids.remove(item_id)
            except exception_type:
                resource_ids.remove(item_id)
            except n_exc.NeutronException as e:
                LOG.error(_LE('Failed to delete %(resource_name)s %(net_id)s '
                              'for service vm due to %(err)s'),
                          {'resource_name': name, 'net_id': item_id, 'err': e})

    def setup_logical_port_connectivity(self, context, port_db,
                                        hosting_device_id):
        pass

    def teardown_logical_port_connectivity(self, context, port_db,
                                           hosting_device_id):
        pass

    def extend_hosting_port_info(self, context, port_db, hosting_device,
                                 hosting_info):
        pass

    def allocate_hosting_port(self, context, router_id, port_db, network_type,
                              hosting_device_id):
        return {'allocated_port_id': None,
                'allocated_vlan': None}



# Copyright 2015 Cisco Systems.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import collections

from oslo_serialization import jsonutils

import mock

from neutron.tests import base

from networking_cisco.apps.saf.agent.vdp import dfa_vdp_mgr
from networking_cisco.apps.saf.agent.vdp import ovs_vdp
from networking_cisco.apps.saf.common import constants
from networking_cisco.apps.saf.common import dfa_sys_lib as utils
from networking_cisco.apps.saf.common import utils as common_utils

try:
    OrderedDict = collections.OrderedDict
except AttributeError:
    import ordereddict
    OrderedDict = ordereddict.OrderedDict


class DfaVdpMgrTest(base.BaseTestCase):
    """A test suite to exercise the DfaVdpMgr Class.  """

    def setUp(self):
        '''Setup routine '''
        super(DfaVdpMgrTest, self).setUp()
        self.root_helper = 'sudo'
        self.port_name = "loc_veth"
        self.uplink = "eth2"
        self.port_str = "loc_veth_eth2"
        self.br_integ = 'br-int1'
        self.br_ex = 'br-ethd1'
        self.rpc_client = mock.Mock()
        self.host = 'hostabc'
        self.config_dict = {'integration_bridge': self.br_integ,
                            'external_bridge': self.br_ex,
                            'root_helper': self.root_helper,
                            'node_list': 'hostabc',
                            'node_uplink_list': self.uplink}
        self.execute = mock.patch.object(
            utils, "execute", spec=utils.execute).start()
        mock.patch('time.sleep').start()
        self.topo_disc = mock.patch(
            'networking_cisco.apps.saf.agent.topo_disc.topo_disc.'
            'TopoDisc').start()
        self._test_dfa_mgr_init()

    def _test_dfa_mgr_init(self):
        '''Test routine for init '''
        with mock.patch('networking_cisco.apps.saf.common.utils.'
                        'EventProcessingThread') as event_fn,\
                mock.patch('networking_cisco.apps.saf.common.'
                           'utils.PeriodicTask') as period_fn:
            event_obj = event_fn.return_value
            period_obj = period_fn.return_value
            parent = mock.MagicMock()
            parent.attach_mock(event_obj.start, 'start')
            parent.attach_mock(period_obj.run, 'run')
            self.dfa_vdp_mgr = dfa_vdp_mgr.VdpMgr(self.config_dict,
                                                  self.rpc_client,
                                                  self.host)
        event_fn.assert_called_with("VDP_Mgr", self.dfa_vdp_mgr,
                                    'process_queue')
        period_fn.assert_any_call(constants.ERR_PROC_INTERVAL,
                                  self.dfa_vdp_mgr.process_err_queue)
        period_fn.assert_any_call(constants.UPLINK_DET_INTERVAL,
                                  self.dfa_vdp_mgr.vdp_uplink_proc_top)
        expected_calls = [mock.call.start(), mock.call.run(), mock.call.run()]
        parent.assert_has_calls(expected_calls)

    def _setup_uplink_msg(self, status):
        '''construct the uplink message '''
        msg = dfa_vdp_mgr.VdpQueMsg(constants.UPLINK_MSG_TYPE, status=status,
                                    phy_uplink=self.uplink,
                                    br_int=self.br_integ,
                                    br_ex=self.br_ex,
                                    root_helper=self.root_helper)
        return msg

    def _setup_vm_msg(self, status):
        '''construct the VM message '''
        msg = dfa_vdp_mgr.VdpQueMsg(constants.VM_MSG_TYPE,
                                    port_uuid='0000-1111-2222-3333',
                                    vm_mac='00:00:fa:11:22:33',
                                    net_uuid='0000-aaaa-bbbb-cccc',
                                    segmentation_id=10001, status=status,
                                    oui=None, phy_uplink=self.uplink)
        return msg

    def _setup_bulk_vm_msg(self, status):
        """construct the Bulk VM message. """
        vm_dict = self._construct_vm_dict(status)
        msg = dfa_vdp_mgr.VdpQueMsg(constants.VM_BULK_SYNC_MSG_TYPE,
                                    vm_bulk_list=[vm_dict],
                                    phy_uplink=self.uplink)
        return msg

    def _construct_vm_dict(self, status):
        """Construct a sample VM dict. """
        vm_dict = {'status': status, 'port_uuid': '0000-1111-2222-3333',
                   'vm_mac': '00:00:fa:11:22:33',
                   'net_uuid': '0000-aaaa-bbbb-cccc', 'vdp_vlan': 3005,
                   'segmentation_id': 10001, 'oui': None, 'local_vlan': 10,
                   'phy_uplink': self.uplink}
        return vm_dict

    def _test_process_up_uplink_event(self):
        '''Test routine when a uplink is detected '''
        msg = self._setup_uplink_msg('up')
        self.dfa_vdp_mgr.phy_uplink = self.uplink
        with mock.patch('networking_cisco.apps.saf.agent.'
                        'vdp.ovs_vdp.OVSNeutronVdp') as ovs_vdp_fn, \
            mock.patch.object(self.dfa_vdp_mgr, 'save_uplink') as (
                save_uplink_fn), \
            mock.patch.object(
                self.dfa_vdp_mgr.topo_disc, 'uncfg_intf') as uncfg_intf_fn, \
            mock.patch.object(
                self.dfa_vdp_mgr.topo_disc, 'cfg_intf') as cfg_intf_fn:
            self.dfa_vdp_mgr.process_uplink_event(msg, self.uplink)
        veth_val = self.dfa_vdp_mgr.ovs_vdp_obj_dict[
            self.uplink].get_lldp_local_bridge_port()
        ovs_vdp_fn.assert_called_with(self.uplink, self.br_integ, self.br_ex,
                                      self.root_helper,
                                      self.dfa_vdp_mgr.vdp_vlan_change_cb)
        save_uplink_fn.assert_called_with(
            uplink=self.uplink, veth_intf=veth_val)
        uncfg_intf_fn.assert_called_with(self.uplink)
        cfg_intf_fn.assert_called_with(
            self.dfa_vdp_mgr.ovs_vdp_obj_dict[
                self.uplink].get_lldp_local_bridge_port(),
            phy_interface=self.uplink)

    def test_process_up_uplink_event_lldp_fail(self):
        """Test routine when a uplink is detected and lldp is down. """
        msg = self._setup_uplink_msg('up')
        self.dfa_vdp_mgr.phy_uplink = self.uplink
        with mock.patch('networking_cisco.apps.saf.agent.'
                        'vdp.ovs_vdp.OVSNeutronVdp') as ovs_vdp_fn, \
            mock.patch.object(self.dfa_vdp_mgr.err_que, 'enqueue') as enq, \
            mock.patch.object(
                self.dfa_vdp_mgr, 'save_uplink') as save_uplink_fn:
            ovs_vdp_fn.return_value.is_lldpad_setup_done.return_value = False
            ovs_vdp_fn.return_value.get_uplink_fail_reason.return_value = (
                'Some Reason')
            self.dfa_vdp_mgr.process_uplink_event(msg, self.uplink)
        ovs_vdp_fn.assert_called_with(self.uplink, self.br_integ, self.br_ex,
                                      self.root_helper,
                                      self.dfa_vdp_mgr.vdp_vlan_change_cb)
        save_uplink_fn.assert_called_with(
            uplink=self.uplink, fail_reason='Some Reason')
        enq.assert_called_with(constants.Q_UPL_PRIO, msg)

    def _test_process_down_uplink_event(self):
        '''Test routine when a uplink goes down '''
        msg = self._setup_uplink_msg('down')
        self.dfa_vdp_mgr.phy_uplink = self.uplink
        del_flow = mock.patch.object(
            ovs_vdp, "delete_uplink_and_flows",
            spec=ovs_vdp.delete_uplink_and_flows).start()
        with mock.patch.object(self.dfa_vdp_mgr, 'save_uplink') as sav_fn, \
            mock.patch.object(
                self.dfa_vdp_mgr.topo_disc, 'uncfg_intf') as uncfg_intf_fn, \
            mock.patch.object(
                self.dfa_vdp_mgr.topo_disc, 'cfg_intf') as cfg_intf_fn:
            parent = mock.MagicMock()
            parent.attach_mock(del_flow, 'delete_uplink_and_flows')
            parent.attach_mock(uncfg_intf_fn, 'uncfg_intf')
            parent.attach_mock(cfg_intf_fn, 'cfg_intf')
            self.dfa_vdp_mgr.process_uplink_event(msg, self.uplink)
        expected_call = [mock.call.delete_uplink_and_flows(self.root_helper,
                                                           self.br_ex,
                                                           self.uplink),
                         mock.call.uncfg_intf(None),
                         mock.call.cfg_intf(self.uplink)]
        parent.assert_has_calls(expected_call)
        sav_fn.assert_called_with()

    def _test_process_down_uplink_event_clear(self):
        '''Test routine when a uplink goes down and flow is cleared '''
        msg = self._setup_uplink_msg('down')
        self.dfa_vdp_mgr.phy_uplink = self.uplink
        self.dfa_vdp_mgr.ovs_vdp_obj_dict = {}
        self.dfa_vdp_mgr.ovs_vdp_obj_dict[self.uplink] = mock.Mock()
        with mock.patch.object(self.dfa_vdp_mgr.ovs_vdp_obj_dict[self.uplink],
                               'clear_obj_params') as vdp_clear_fn, \
            mock.patch.object(self.dfa_vdp_mgr, 'save_uplink') as sav_fn, \
            mock.patch.object(
                self.dfa_vdp_mgr.topo_disc, 'uncfg_intf') as uncfg_intf_fn, \
            mock.patch.object(
                self.dfa_vdp_mgr.topo_disc, 'cfg_intf') as cfg_intf_fn:
            self.dfa_vdp_mgr.process_uplink_event(msg, self.uplink)
        vdp_clear_fn.assert_called_with()
        sav_fn.assert_called_with()
        uncfg_intf_fn.assert_called_with(None)
        cfg_intf_fn.assert_called_with(self.uplink)

    def _test_process_vm_event_succ(self):
        '''Test routine for VM event process is successful '''
        msg = self._setup_vm_msg('up')
        lvid = 10
        vdp_vlan = 3000
        self.dfa_vdp_mgr.uplink_det_compl = True
        port_reply = {'result': True, 'fail_reason': None}
        self.dfa_vdp_mgr.ovs_vdp_obj_dict = {self.uplink: mock.Mock()}
        with mock.patch.object(
            self.dfa_vdp_mgr, 'update_vm_result') as save_vmres_fn, \
            mock.patch.object(self.dfa_vdp_mgr.ovs_vdp_obj_dict[self.uplink],
                              'send_vdp_port_event',
                              return_value=port_reply), \
            mock.patch.object(self.dfa_vdp_mgr.ovs_vdp_obj_dict[self.uplink],
                              'get_lvid_vdp_vlan',
                              return_value=[lvid, vdp_vlan]):
            self.dfa_vdp_mgr.process_vm_event(msg, self.uplink)
        save_vmres_fn.assert_called_with('0000-1111-2222-3333', 'SUCCESS',
                                         fail_reason=None, lvid=lvid,
                                         vdp_vlan=vdp_vlan)

    def _test_process_vm_event_fail(self):
        '''Test routine for VM event process has failed '''
        msg = self._setup_vm_msg('up')
        self.dfa_vdp_mgr.uplink_det_compl = True
        self.dfa_vdp_mgr.ovs_vdp_obj_dict = {self.uplink: mock.Mock()}
        port_reply = {'result': False, 'fail_reason': 'some reason'}
        with mock.patch.object(
            self.dfa_vdp_mgr, 'update_vm_result') as save_vmres_fn, \
            mock.patch.object(self.dfa_vdp_mgr.ovs_vdp_obj_dict[self.uplink],
                              'send_vdp_port_event',
                              return_value=port_reply):
            self.dfa_vdp_mgr.process_vm_event(msg, self.uplink)
        save_vmres_fn.assert_called_with('0000-1111-2222-3333', 'CREATE:FAIL',
                                         fail_reason='some reason')

    def test_process_vm_event_uplink_not_rcvd(self):
        """Test for VM event process, when uplink event is not received. """
        msg = self._setup_vm_msg('up')
        self.dfa_vdp_mgr.uplink_det_compl = True
        with mock.patch.object(self.dfa_vdp_mgr, 'update_vm_result') as \
                save_vmres_fn:
            self.dfa_vdp_mgr.process_vm_event(msg, self.uplink)
        save_vmres_fn.assert_called_with('0000-1111-2222-3333', 'CREATE:FAIL')

    def test_process_vm_down_event_uplink_not_rcvd(self):
        """Test for VM down event, when uplink event is not received. """
        msg = self._setup_vm_msg('down')
        self.dfa_vdp_mgr.uplink_det_compl = True
        with mock.patch.object(self.dfa_vdp_mgr, 'update_vm_result') as \
                save_vmres_fn:
            self.dfa_vdp_mgr.process_vm_event(msg, self.uplink)
        save_vmres_fn.assert_called_with('0000-1111-2222-3333', 'DELETE:FAIL')

    def test_process_uplink_event_case1(self):
        '''Top routine that calls the uplink down case '''
        self._test_process_down_uplink_event()

    def test_process_uplink_event_case2(self):
        '''Top routine that calls the uplink detect case '''
        self._test_process_up_uplink_event()
        self._test_process_down_uplink_event_clear()

    def test_process_vm_event_succ(self):
        '''Top routine that calls process VM event success case '''
        self._test_process_vm_event_succ()

    def test_process_vm_event_fail(self):
        '''Top routine that calls process VM event fail case '''
        self._test_process_vm_event_fail()

    def test_process_static_uplink_new(self):
        """Test routine for static uplink, first time call."""
        ret = self.dfa_vdp_mgr.static_uplink_detect(None)
        self.assertEqual(ret, self.uplink)

    def test_process_static_uplink_normal(self):
        """Test routine for static uplink, normal case."""
        ret = self.dfa_vdp_mgr.static_uplink_detect('veth_temp')
        self.assertEqual(ret, 'normal')

    def test_process_static_uplink_down(self):
        """Test routine for static uplink, down case."""
        self.dfa_vdp_mgr.phy_uplink = 'eth3'
        ret = self.dfa_vdp_mgr.static_uplink_detect(None)
        self.assertEqual(ret, 'down')

    def test_vdp_vlan_change_cb(self):
        """Function to test the VDP VLAN change Callback. """
        port_uuid = '0000-1111-2222-3333'
        lvid = 10
        vdp_vlan = 3000
        with mock.patch.object(self.dfa_vdp_mgr, 'update_vm_result') as \
                save_vmres_fn:
            self.dfa_vdp_mgr.vdp_vlan_change_cb(port_uuid, lvid, vdp_vlan,
                                                'some reason')
        save_vmres_fn.assert_called_with('0000-1111-2222-3333', 'SUCCESS',
                                         lvid=lvid, vdp_vlan=vdp_vlan,
                                         fail_reason='some reason')

    def _test_setup_uplink_params_steady(self):
        """Internal function to setup the DFA VDP Mgr parameters. """
        self.dfa_vdp_mgr.is_os_run = True
        self.dfa_vdp_mgr.restart_uplink_called = True
        self.dfa_vdp_mgr.process_uplink_ongoing = False
        self.dfa_vdp_mgr.uplink_det_compl = False
        self.dfa_vdp_mgr.static_uplink = False
        self.dfa_vdp_mgr.phy_uplink = self.uplink
        self.dfa_vdp_mgr.ovs_vdp_obj_dict = {self.uplink: mock.Mock()}

    def test_vdp_uplink_proc_down_threshold_exceed(self):
        """Test VDP uplink proc for down case, when threshold exceed. """
        self._test_setup_uplink_params_steady()
        self.dfa_vdp_mgr.uplink_down_cnt = 3
        with mock.patch('networking_cisco.apps.saf.agent.detect_uplink.'
                        'detect_uplink', return_value='down'), \
            mock.patch('networking_cisco.apps.saf.agent.vdp.dfa_vdp_mgr.'
                       'VdpQueMsg', return_value='some obj') as vdp_que, \
            mock.patch.object(self.dfa_vdp_mgr.que, 'enqueue') as enq, \
            mock.patch.object(self.dfa_vdp_mgr.ovs_vdp_obj_dict[self.uplink],
                              'get_lldp_local_bridge_port',
                              return_value='veth_eth1'):
            self.dfa_vdp_mgr.vdp_uplink_proc()
        vdp_que.assert_called_with(
            constants.UPLINK_MSG_TYPE, status='down',
            phy_uplink=self.uplink, br_int=self.dfa_vdp_mgr.br_integ,
            br_ex=self.dfa_vdp_mgr.br_ex,
            root_helper=self.dfa_vdp_mgr.root_helper)
        enq.assert_called_with(constants.Q_UPL_PRIO, 'some obj')

    def test_vdp_uplink_proc_down_threshold_not_exceed(self):
        """Test VDP uplink proc for down case, when threshold not exceed. """
        self._test_setup_uplink_params_steady()
        self.dfa_vdp_mgr.uplink_down_cnt = 1
        with mock.patch('networking_cisco.apps.saf.agent.detect_uplink.'
                        'detect_uplink', return_value='down'), \
            mock.patch('networking_cisco.apps.saf.agent.vdp.dfa_vdp_mgr.'
                       'VdpQueMsg', return_value='some obj') as vdp_que, \
            mock.patch.object(self.dfa_vdp_mgr.que, 'enqueue') as enq, \
            mock.patch.object(self.dfa_vdp_mgr.ovs_vdp_obj_dict[self.uplink],
                              'get_lldp_local_bridge_port',
                              return_value='veth_eth1'):
            self.dfa_vdp_mgr.vdp_uplink_proc()
        vdp_que.assert_not_called()
        enq.assert_not_called()

    def test_vdp_uplink_proc_none(self):
        """Test VDP uplink proc for none case. """
        self._test_setup_uplink_params_steady()
        with mock.patch('networking_cisco.apps.saf.agent.detect_uplink.'
                        'detect_uplink', return_value=None), \
            mock.patch('networking_cisco.apps.saf.agent.vdp.dfa_vdp_mgr.'
                       'VdpQueMsg', return_value='some obj') as vdp_que, \
            mock.patch.object(self.dfa_vdp_mgr.ovs_vdp_obj_dict[self.uplink],
                              'get_lldp_local_bridge_port',
                              return_value=None), \
            mock.patch.object(self.dfa_vdp_mgr.que, 'enqueue') as enq, \
            mock.patch.object(self.dfa_vdp_mgr, 'save_uplink') as (
                save_uplink_fn):
            self.dfa_vdp_mgr.vdp_uplink_proc()
        vdp_que.assert_not_called()
        enq.assert_not_called()
        save_uplink_fn.assert_called_with(
            fail_reason=constants.uplink_undiscovered_reason)

    def test_vdp_uplink_proc_normal(self):
        """Test VDP uplink proc for normal case. """
        self._test_setup_uplink_params_steady()
        with mock.patch('networking_cisco.apps.saf.agent.detect_uplink.'
                        'detect_uplink', return_value='normal'), \
            mock.patch('networking_cisco.apps.saf.agent.vdp.dfa_vdp_mgr.'
                       'VdpQueMsg', return_value='some obj') as vdp_que, \
            mock.patch('networking_cisco.apps.saf.common.dfa_sys_lib.'
                       'get_bond_intf', return_value=None), \
            mock.patch.object(self.dfa_vdp_mgr.ovs_vdp_obj_dict[self.uplink],
                              'get_lldp_local_bridge_port',
                              return_value='veth_eth1'), \
            mock.patch.object(self.dfa_vdp_mgr.que, 'enqueue') as enq, \
            mock.patch.object(self.dfa_vdp_mgr, 'save_uplink') as (
                save_uplink_fn):
            self.dfa_vdp_mgr.vdp_uplink_proc()
        vdp_que.assert_called_with(
            constants.UPLINK_MSG_TYPE, status='up',
            phy_uplink=self.uplink, br_int=self.dfa_vdp_mgr.br_integ,
            br_ex=self.dfa_vdp_mgr.br_ex,
            root_helper=self.dfa_vdp_mgr.root_helper)
        enq.assert_called_with(constants.Q_UPL_PRIO, 'some obj')
        save_uplink_fn.assert_not_called()

    def test_vdp_uplink_proc_normal_static(self):
        """Test VDP uplink proc for normal case for static uplink. """
        self._test_setup_uplink_params_steady()
        self.dfa_vdp_mgr.static_uplink = True
        with mock.patch('networking_cisco.apps.saf.agent.vdp.dfa_vdp_mgr.'
                        'VdpQueMsg', return_value='some obj') as vdp_que, \
            mock.patch('networking_cisco.apps.saf.common.dfa_sys_lib.'
                       'get_bond_intf', return_value=None), \
            mock.patch.object(self.dfa_vdp_mgr, 'static_uplink_detect',
                              return_value='normal'), \
            mock.patch.object(self.dfa_vdp_mgr.ovs_vdp_obj_dict[self.uplink],
                              'get_lldp_local_bridge_port',
                              return_value='veth_eth1'), \
            mock.patch.object(self.dfa_vdp_mgr.que, 'enqueue') as enq, \
            mock.patch.object(self.dfa_vdp_mgr, 'save_uplink') as (
                save_uplink_fn):
            self.dfa_vdp_mgr.vdp_uplink_proc()
        vdp_que.assert_called_with(
            constants.UPLINK_MSG_TYPE, status='up',
            phy_uplink=self.uplink, br_int=self.dfa_vdp_mgr.br_integ,
            br_ex=self.dfa_vdp_mgr.br_ex,
            root_helper=self.dfa_vdp_mgr.root_helper)
        enq.assert_called_with(constants.Q_UPL_PRIO, 'some obj')
        save_uplink_fn.assert_not_called()

    def test_vdp_uplink_proc_normal_bulk_vm_not_rcvd(self):
        """Test VDP uplink proc for normal case, for bulk VM.

        This is for the case when bulk VM notificationis not received.
        """
        self._test_setup_uplink_params_steady()
        self.dfa_vdp_mgr.bulk_vm_rcvd_flag = False
        self.dfa_vdp_mgr.bulk_vm_check_cnt = 2
        with mock.patch('networking_cisco.apps.saf.agent.detect_uplink.'
                        'detect_uplink', return_value='normal'), \
            mock.patch('networking_cisco.apps.saf.agent.vdp.dfa_vdp_mgr.'
                       'VdpQueMsg', return_value='some obj') as vdp_que, \
            mock.patch('networking_cisco.apps.saf.common.dfa_sys_lib.'
                       'get_bond_intf', return_value=None), \
            mock.patch.object(self.dfa_vdp_mgr.ovs_vdp_obj_dict[self.uplink],
                              'get_lldp_local_bridge_port',
                              return_value='veth_eth1'), \
            mock.patch.object(self.dfa_vdp_mgr.que, 'enqueue') as enq, \
            mock.patch.object(self.dfa_vdp_mgr, 'save_uplink') as (
                save_uplink_fn):
            self.dfa_vdp_mgr.vdp_uplink_proc()
        vdp_que.assert_called_with(
            constants.UPLINK_MSG_TYPE, status='up',
            phy_uplink=self.uplink, br_int=self.dfa_vdp_mgr.br_integ,
            br_ex=self.dfa_vdp_mgr.br_ex,
            root_helper=self.dfa_vdp_mgr.root_helper)
        enq.assert_called_with(constants.Q_UPL_PRIO, 'some obj')
        save_uplink_fn.assert_called_with(uplink=self.uplink,
                                          veth_intf='veth_eth1')

    def test_vdp_uplink_proc_normal_bond_intf(self):
        """Test VDP uplink proc for normal case, for bond interface. """
        self._test_setup_uplink_params_steady()
        with mock.patch('networking_cisco.apps.saf.agent.detect_uplink.'
                        'detect_uplink', return_value='normal'), \
            mock.patch('networking_cisco.apps.saf.agent.vdp.dfa_vdp_mgr.'
                       'VdpQueMsg', return_value='some obj') as vdp_que, \
            mock.patch('networking_cisco.apps.saf.common.dfa_sys_lib.'
                       'get_bond_intf', return_value='bond0'), \
            mock.patch.object(self.dfa_vdp_mgr.ovs_vdp_obj_dict[self.uplink],
                              'get_lldp_local_bridge_port',
                              return_value='veth_eth1'), \
            mock.patch.object(self.dfa_vdp_mgr.que, 'enqueue') as enq, \
            mock.patch.object(self.dfa_vdp_mgr, 'save_uplink') as (
                save_uplink_fn):
            parent = mock.MagicMock()
            parent.attach_mock(save_uplink_fn, 'save_uplink')
            parent.attach_mock(vdp_que, 'VdpQueMsg')
            parent.attach_mock(enq, 'enqueue')
            self.dfa_vdp_mgr.vdp_uplink_proc()
        expected_calls = [
            mock.call.save_uplink(
                fail_reason=constants.port_transition_bond_down_reason),
            mock.call.VdpQueMsg(
                constants.UPLINK_MSG_TYPE, status='down',
                phy_uplink=self.uplink, br_int=self.dfa_vdp_mgr.br_integ,
                br_ex=self.dfa_vdp_mgr.br_ex,
                root_helper=self.dfa_vdp_mgr.root_helper),
            mock.call.enqueue(1, 'some obj'),
            mock.call.save_uplink(
                fail_reason=constants.port_transition_bond_up_reason,
                uplink='bond0'),
            mock.call.VdpQueMsg(
                constants.UPLINK_MSG_TYPE, status='up',
                phy_uplink='bond0', br_int=self.dfa_vdp_mgr.br_integ,
                br_ex=self.dfa_vdp_mgr.br_ex,
                root_helper=self.dfa_vdp_mgr.root_helper),
            mock.call.enqueue(1, 'some obj')]
        parent.assert_has_calls(expected_calls, any_order=False)
        self.assertEqual('bond0', self.dfa_vdp_mgr.phy_uplink)
        self.assertTrue(self.dfa_vdp_mgr.process_uplink_ongoing)
        self.assertFalse(self.dfa_vdp_mgr.uplink_det_compl)

    def test_vdp_uplink_proc_new_uplink(self):
        """Test VDP uplink proc for a new uplink case. """
        self._test_setup_uplink_params_steady()
        with mock.patch('networking_cisco.apps.saf.agent.detect_uplink.'
                        'detect_uplink', return_value='eth4'), \
            mock.patch('networking_cisco.apps.saf.agent.vdp.dfa_vdp_mgr.'
                       'VdpQueMsg', return_value='some obj') as vdp_que, \
            mock.patch.object(self.dfa_vdp_mgr.ovs_vdp_obj_dict[self.uplink],
                              'get_lldp_local_bridge_port',
                              return_value='veth_eth1'), \
            mock.patch.object(self.dfa_vdp_mgr.que, 'enqueue') as enq, \
            mock.patch.object(self.dfa_vdp_mgr, 'save_uplink') as (
                save_uplink_fn):
            self.dfa_vdp_mgr.vdp_uplink_proc()
        vdp_que.assert_called_with(
            constants.UPLINK_MSG_TYPE, status='up',
            phy_uplink='eth4', br_int=self.dfa_vdp_mgr.br_integ,
            br_ex=self.dfa_vdp_mgr.br_ex,
            root_helper=self.dfa_vdp_mgr.root_helper)
        enq.assert_called_with(constants.Q_UPL_PRIO, 'some obj')
        save_uplink_fn.assert_called_with(uplink='eth4', veth_intf='veth_eth1')
        self.assertTrue(self.dfa_vdp_mgr.process_uplink_ongoing)
        self.assertEqual('eth4', self.dfa_vdp_mgr.phy_uplink)

    def test_process_bulk_vm_event_up(self):
        "Test Bulk Process event for up status. """
        msg = self._setup_bulk_vm_msg('up')
        self.dfa_vdp_mgr.ovs_vdp_obj_dict = {}
        self.dfa_vdp_mgr.ovs_vdp_obj_dict[self.uplink] = mock.Mock()
        with mock.patch('networking_cisco.apps.saf.agent.vdp.dfa_vdp_mgr.'
                        'VdpQueMsg', return_value='some obj') as vdp_que, \
            mock.patch.object(self.dfa_vdp_mgr,
                              'process_vm_event') as process_vm_event:
            self.dfa_vdp_mgr.uplink_det_compl = True
            self.dfa_vdp_mgr.process_bulk_vm_event(msg, self.uplink)
        vm_dict = self._construct_vm_dict('up')
        vdp_que.assert_called_with(
            constants.VM_MSG_TYPE, port_uuid=vm_dict['port_uuid'],
            vm_mac=vm_dict['vm_mac'], net_uuid=vm_dict['net_uuid'],
            segmentation_id=vm_dict['segmentation_id'],
            status=vm_dict['status'], oui=vm_dict['oui'],
            phy_uplink=self.uplink)
        process_vm_event.assert_called_with('some obj', self.uplink)

    def test_process_bulk_vm_event_down(self):
        "Test Bulk Process event for down status. """
        msg = self._setup_bulk_vm_msg('down')
        self.dfa_vdp_mgr.ovs_vdp_obj_dict = {}
        self.dfa_vdp_mgr.ovs_vdp_obj_dict[self.uplink] = mock.Mock()
        with mock.patch('networking_cisco.apps.saf.agent.vdp.dfa_vdp_mgr.'
                        'VdpQueMsg', return_value='some obj') as vdp_que, \
            mock.patch.object(self.dfa_vdp_mgr.ovs_vdp_obj_dict[self.uplink],
                              'pop_local_cache') as pop_ovs_cache, \
            mock.patch.object(self.dfa_vdp_mgr,
                              'process_vm_event') as process_vm_event:
            self.dfa_vdp_mgr.uplink_det_compl = True
            self.dfa_vdp_mgr.process_bulk_vm_event(msg, self.uplink)
        vm_dict = self._construct_vm_dict('down')
        pop_ovs_cache.assert_called_with(
            vm_dict['port_uuid'], vm_dict['vm_mac'], vm_dict['net_uuid'],
            vm_dict['local_vlan'], vm_dict['vdp_vlan'],
            vm_dict['segmentation_id'])
        vdp_que.assert_called_with(
            constants.VM_MSG_TYPE, port_uuid=vm_dict['port_uuid'],
            vm_mac=vm_dict['vm_mac'], net_uuid=vm_dict['net_uuid'],
            segmentation_id=vm_dict['segmentation_id'],
            status=vm_dict['status'], oui=vm_dict['oui'],
            phy_uplink=self.uplink)
        process_vm_event.assert_called_with('some obj', self.uplink)

    def test_vdp_vm_event_dict(self):
        """Test for VM event, when input is a VM dict. """
        vm_dict = self._construct_vm_dict('up')
        self.dfa_vdp_mgr.phy_uplink = self.uplink
        with mock.patch('networking_cisco.apps.saf.agent.vdp.dfa_vdp_mgr.'
                        'VdpQueMsg', return_value='some obj') as vdp_que, \
            mock.patch.object(self.dfa_vdp_mgr.que, 'enqueue') as enq, \
            mock.patch('networking_cisco.apps.saf.agent.vdp.dfa_vdp_mgr.'
                       'VdpQueMsg', return_value='some obj') as vdp_que:
            self.dfa_vdp_mgr.vdp_vm_event(vm_dict)
        vdp_que.assert_called_with(
            constants.VM_MSG_TYPE, port_uuid=vm_dict['port_uuid'],
            vm_mac=vm_dict['vm_mac'], net_uuid=vm_dict['net_uuid'],
            segmentation_id=vm_dict['segmentation_id'],
            status=vm_dict['status'], oui=vm_dict['oui'],
            phy_uplink=self.uplink)
        enq.assert_called_with(constants.Q_VM_PRIO, 'some obj')

    def test_vdp_vm_event_list(self):
        """Test for VM event, when input is a VM dict list. """
        vm_dict = self._construct_vm_dict('up')
        self.dfa_vdp_mgr.phy_uplink = self.uplink
        with mock.patch('networking_cisco.apps.saf.agent.vdp.dfa_vdp_mgr.'
                        'VdpQueMsg', return_value='some obj') as vdp_que, \
            mock.patch.object(self.dfa_vdp_mgr.que, 'enqueue') as enq, \
            mock.patch('networking_cisco.apps.saf.agent.vdp.dfa_vdp_mgr.'
                       'VdpQueMsg', return_value='some obj') as vdp_que:
            self.dfa_vdp_mgr.vdp_vm_event([vm_dict])
        vdp_que.assert_called_with(
            constants.VM_BULK_SYNC_MSG_TYPE,
            vm_bulk_list=[vm_dict], phy_uplink=self.uplink)
        enq.assert_called_with(constants.Q_VM_PRIO, 'some obj')
        self.assertTrue(self.dfa_vdp_mgr.bulk_vm_rcvd_flag)

    def test_dfa_uplink_restart_no_uplink(self):
        """Test for DFA uplink restart case with no uplink. """
        uplink_dict = {}
        with mock.patch('networking_cisco.apps.saf.agent.vdp.ovs_vdp.'
                        'delete_uplink_and_flows') as ovs_del_flow:
            self.dfa_vdp_mgr.dfa_uplink_restart(uplink_dict)
        self.assertTrue(self.dfa_vdp_mgr.restart_uplink_called)
        ovs_del_flow.assert_not_called()

    def test_dfa_uplink_restart_valid_uplink_veth(self):
        """Test for DFA uplink restart case with valid uplink/veth. """
        uplink_dict = {'uplink': self.uplink, 'veth_intf': 'veth_eth1'}
        with mock.patch('networking_cisco.apps.saf.agent.vdp.ovs_vdp.'
                        'delete_uplink_and_flows') as ovs_del_flow:
            self.dfa_vdp_mgr.dfa_uplink_restart(uplink_dict)
        self.assertTrue(self.dfa_vdp_mgr.restart_uplink_called)
        self.assertEqual(self.uplink, self.dfa_vdp_mgr.phy_uplink)
        self.assertEqual('veth_eth1', self.dfa_vdp_mgr.veth_intf)
        ovs_del_flow.assert_not_called()

    def test_dfa_uplink_restart_invalid_veth(self):
        """Test for DFA uplink restart case with invalid veth. """
        uplink_dict = {'uplink': self.uplink}
        with mock.patch('networking_cisco.apps.saf.agent.vdp.ovs_vdp.'
                        'delete_uplink_and_flows') as ovs_del_flow:
            self.dfa_vdp_mgr.dfa_uplink_restart(uplink_dict)
        self.assertTrue(self.dfa_vdp_mgr.restart_uplink_called)
        ovs_del_flow.assert_called_with(
            self.dfa_vdp_mgr.root_helper, self.dfa_vdp_mgr.br_ex, self.uplink)

    def _get_topo_disc_arg(self):
        """Fill the argument to topology discovery. """
        return {'phy_interface': self.uplink, 'remote_evb_cfgd': True,
                'remote_evb_mode': 'bridge', 'remote_mgmt_addr': '10.1.1.1',
                'remote_system_desc': 'Cisco Test Os',
                'remote_system_name': 'N6K-1', 'remote_port': 'e2/1',
                'remote_chassis_id_mac': '00:11:22:33:44:55',
                'remote_port_id_mac': '00:22:33:44:55:66'}

    def _get_topo_dict_final(self):
        """Fill the argument to topology discovery for final check. """
        return {'host': None, 'protocol_interface': self.uplink,
                'heartbeat': 'Jan 1',
                'phy_interface': self.uplink, 'remote_evb_cfgd': True,
                'remote_evb_mode': 'bridge',
                'remote_mgmt_addr': '10.1.1.1',
                'remote_system_desc': 'Cisco Test Os',
                'remote_system_name': 'N6K-1', 'remote_port': 'e2/1',
                'remote_chassis_id_mac': '00:11:22:33:44:55',
                'remote_port_id_mac': '00:22:33:44:55:66',
                'configurations': jsonutils.dumps({})}

    def test_topo_disc_cb(self):
        """Test the topology discovery CB function. """
        topo_dict = self._get_topo_disc_arg()
        topo_dict_final = self._get_topo_dict_final()
        topo_obj = common_utils.Dict2Obj(topo_dict)
        with mock.patch('time.ctime') as ctime_fn:
            ctime_fn.return_value = 'Jan 1'
            self.dfa_vdp_mgr.topo_disc_cb(self.uplink, topo_obj)
        self.rpc_client.make_msg.assert_called_with(
            'save_topo_disc_params', {}, msg=jsonutils.dumps(topo_dict_final))



# Copyright (c) 2014-2016 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg
import six

from neutron.tests.unit import testlib_api

from networking_cisco.tests import base as nc_base

test_config_file = """
[ml2_mech_cisco_nexus:1.1.1.1]
username=admin
password=mySecretPassword
ssh_port=22
nve_src_intf=2
physnet=physnet1
vpc_pool=5,10
intfcfg_portchannel=user cmd1;user cmd2
https_verify=True
https_local_certificate=/path/to/your/local-certificate-file.crt
compute1=1/1
compute2=1/2
compute5=1/3,1/4

[ml2_mech_cisco_nexus:2.2.2.2]
username=admin
password=mySecretPassword
ssh_port=22
compute3=1/1
compute4=1/2
compute5=portchannel:20,portchannel:30
"""

# Make sure intfcfg.portchannel still works
test_deprecate_config_file = """
[ml2_mech_cisco_nexus:1.1.1.1]
username=admin
password=mySecretPassword
ssh_port=22
nve_src_intf=2
physnet=physnet1
vpc_pool=5,10
intfcfg.portchannel=user cmd1;user cmd2
https_verify=True
https_local_certificate=/path/to/your/local-certificate-file.crt
compute1=1/1
compute2=1/2
compute5=1/3,1/4
"""

# Assign non-integer to ssh_port for error
test_error_config_file = """
[ml2_mech_cisco_nexus:1.1.1.1]
username=admin
password=mySecretPassword
ssh_port='abc'
nve_src_intf=2
physnet=physnet1
compute1=1/1
"""

# Assign non-integer to ssh_port for error
dict_mapping_config_file = """
[ml2_mech_cisco_nexus:1.1.1.1]
username=admin
password=mySecretPassword
nve_src_intf=2
physnet=physnet1
host_ports_mapping=compute1:[1/1],
                   compute2:[1/2],
                   compute3:[1/3, port-channel30]
"""


class TestCiscoNexusPluginConfigBase(testlib_api.SqlTestCase):

    def setUp(self):
        super(TestCiscoNexusPluginConfigBase, self).setUp()
        cfg.CONF.clear()


class TestCiscoNexusPluginConfig(TestCiscoNexusPluginConfigBase):

    def test_config_using_subsection_option(self):
        nc_base.load_config_file(test_config_file)
        expected = {
            '1.1.1.1': {
                'username': 'admin',
                'password': 'mySecretPassword',
                'ssh_port': 22,
                'nve_src_intf': '2',
                'physnet': 'physnet1',
                'vpc_pool': '5,10',
                'intfcfg_portchannel': 'user cmd1;user cmd2',
                'https_verify': True,
                'https_local_certificate': (
                    '/path/to/your/local-certificate-file.crt'),
                'host_port_mapping': {
                    'compute1': '1/1',
                    'compute2': '1/2',
                    'compute5': '1/3,1/4'
                }
            }, '2.2.2.2': {
                'username': 'admin',
                'password': 'mySecretPassword',
                'ssh_port': 22,
                'physnet': None,
                'nve_src_intf': None,
                'vpc_pool': None,
                'intfcfg_portchannel': None,
                'https_verify': False,
                'https_local_certificate': None,
                'host_port_mapping': {
                    'compute3': '1/1',
                    'compute4': '1/2',
                    'compute5': 'portchannel:20,portchannel:30'
                }
            }
        }

        for switch_ip, options in expected.items():
            for opt_name, option in options.items():
                self.assertEqual(
                    option, cfg.CONF.ml2_cisco.nexus_switches.get(
                        switch_ip).get(opt_name))

    def test_deprecated_intfcfg_portchannel(self):
        nc_base.load_config_file(test_deprecate_config_file)
        """Test creation deprecated intfcfg_portchannel works."""
        expected = {
            '1.1.1.1': {
                'username': 'admin',
                'password': 'mySecretPassword',
                'ssh_port': 22,
                'nve_src_intf': '2',
                'physnet': 'physnet1',
                'vpc_pool': '5,10',
                'intfcfg_portchannel': 'user cmd1;user cmd2',
                'https_verify': True,
                'https_local_certificate': (
                    '/path/to/your/local-certificate-file.crt'),
                'host_port_mapping': {
                    'compute1': '1/1',
                    'compute2': '1/2',
                    'compute5': '1/3,1/4'
                }
            }
        }

        for switch_ip, options in expected.items():
            for opt_name, option in options.items():
                self.assertEqual(
                    option, cfg.CONF.ml2_cisco.nexus_switches.get(
                        switch_ip).get(opt_name))

    def test_create_device_error(self):
        nc_base.load_config_file(test_error_config_file)
        """Test error during create of the Nexus device dictionary."""

        e = self.assertRaises(cfg.ConfigFileValueError,
            cfg.CONF.ml2_cisco.nexus_switches.get('1.1.1.1').get,
            "ssh_port")
        x = six.u(str(e))
        self.assertIn("Value for option ssh_port is not valid: "
                      "invalid literal for int() with base 10: "
                      "'abc'", x)

    def test_dict_host_port_mapping(self):
        nc_base.load_config_file(dict_mapping_config_file)
        """Test port_host_mapping dictionary works."""
        expected = {
            '1.1.1.1': {
                'username': 'admin',
                'password': 'mySecretPassword',
                'host_ports_mapping': {
                    'compute1': ['1/1'],
                    'compute2': ['1/2'],
                    'compute3': ['1/3', 'port-channel30']
                }
            }
        }

        for switch_ip, options in expected.items():
            for opt_name, option in options.items():
                self.assertEqual(
                    option, cfg.CONF.ml2_cisco.nexus_switches.get(
                        switch_ip).get(opt_name))
